import logging
import subprocess
import os
import shutil
import sys

from sfaira.commands.questionary import sfaira_questionary
from rich import print


log = logging.getLogger(__name__)


class PullRequestHandler:

    def __init__(self, path_loader):
        self.WD = os.path.dirname(__file__)
        self.path_loader = path_loader
        self.loader_name_list = []
        self.loader_name = None
        self.gh_token = ""

    def submit_pr(self):
        self._check_container_and_loaders()
        self._gh_authenticate()
        self._clone_sfaira_and_move_dataloader()
        self._submit_pr()

    def _check_container_and_loaders(self):
        sfaira_container = os.getenv('SFAIRA_DOCKER')
        if sfaira_container and self.path_loader and os.path.isdir(self.path_loader):
            for d in os.listdir(self.path_loader):
                if os.path.isdir(os.path.join(self.path_loader, d)) and d.startswith(("d10_", "dno_doi_")):
                    self.loader_name_list.append(d)
            if not self.loader_name_list:
                print("[bold red]No data loaders found in loader directory. Aborting.")
                sys.exit()
        else:
            print('[bold red]This function can only be called when running inside the sfaira CLI docker container. '
                  'Aborting.')
            sys.exit()

    def _gh_authenticate(self) -> None:
        """
        Guides the user to authenticate with the GitHub CLI
        """
        # Skip login procedure if already logged in
        if subprocess.run(["gh", "auth", "status"], check=False, text=True, shell=False).returncode == 0:
            subprocess.run(["gh", "auth", "setup-git"], check=True, text=True, shell=False)
            print("[bold green]Already authenticated with GitHub.")
            return
        print(
            "[bold blue]Please authenticate with GitHub. Hint: you can generate a Personal Access Token here: "
            "https://github.com/settings/tokens\nThe minimum required scopes are 'repo', 'read:org', 'workflow'."
        )
        returncode = 1
        while returncode != 0:
            gh_token = sfaira_questionary(function='password',
                                          question='Please enter your GitHub token. '
                                                   'Leave blank to interactively authenticate with GitHub.',
                                          default='')
            if gh_token == "":
                gh_call = subprocess.run(
                    ["gh", "auth", "login", "--hostname", "github.com"], check=False, text=True, shell=False)
            else:
                gh_call = subprocess.run(
                    f"echo {gh_token} | gh auth login --with-token", check=False, text=True, shell=True)
            returncode = gh_call.returncode
        self.gh_token = gh_token
        subprocess.run(["gh", "auth", "setup-git"], check=True, text=True, shell=False)
        print("[bold green]Successfully authenticated with GitHub.")

    def _clone_sfaira_and_move_dataloader(self) -> None:
        """
        Clones the sfaira repo, creates a new branch and moves dataloader into the right location.
        """
        # Clone sfaira
        subprocess.run(["rm", "-rf", "/root/sfaira"], check=False, text=True, shell=False)
        subprocess.run(
            ["gh", "repo", "clone", "theislab/sfaira", "/root/sfaira/"], check=True, text=True, shell=False)
        # Get loader name
        if len(self.loader_name_list) == 1:
            self.loader_name = self.loader_name_list[0]
        else:
            self.loader_name = sfaira_questionary(
                function='select',
                question='Multiple dataloaders detected in the loader directory. '
                         'Which one do you want to submit as a pull request?',
                choices=self.loader_name_list
            )
        # Create new branch in sfaira git repo
        subprocess.run(
            ["git", "checkout", "-b", f"dataset/{self.loader_name}"],
            check=True, text=True, shell=False, cwd="/root/sfaira/"
        )
        # Copy loader
        shutil.copytree(
            src=os.path.join(self.path_loader, self.loader_name),
            dst=os.path.join("/root/sfaira/data/dataloaders/loaders", self.loader_name)
        )

    def _submit_pr(self):
        """
        Adds and commits the dataloader in git and creates a pull request
        """
        # Make sure git credentials are set and get them from github api if not
        current_email = subprocess.run(["git", "config", "user.email"], check=False, text=True, shell=False,
                                       stdout=subprocess.PIPE).stdout
        if not current_email:
            git_email = subprocess.run(["gh", "api", "/user/public_emails", "-q", ".[0].email"], check=True, text=True,
                                       shell=False, stdout=subprocess.PIPE).stdout.strip()
            subprocess.run(["git", "config", "--global", "user.email", git_email], check=True, text=True, shell=False)
        current_user = subprocess.run(["git", "config", "user.name"], check=False, text=True, shell=False,
                                      stdout=subprocess.PIPE).stdout
        if not current_user:
            git_user = subprocess.run(["gh", "api", "/user", "-q", ".login"], check=True, text=True, shell=False,
                                      stdout=subprocess.PIPE).stdout.strip()
            subprocess.run(["git", "config", "--global", "user.name", git_user], check=True, text=True, shell=False)
        # Add and commit the dataloader
        subprocess.run(["git", "add", "*"], check=True, text=True, shell=False, cwd="/root/sfaira/")
        subprocess.run(["git", "commit", "-m", f"[from sfaira cli] add dataloader {self.loader_name}"],
                       check=True, text=True, shell=False, cwd="/root/sfaira/")
        # Create pullrequest (authenticate again beforehand if gh_token was provided before)
        create_pr_str = f"gh pr create --base dev --title {self.loader_name} " \
                        f"--body 'This PR was created by the sfaira CLI adding dataset {self.loader_name}' " \
                        f"--label dataset"
        if self.gh_token != "":
            subprocess.run(f"echo {self.gh_token} | gh auth login --with-token && {create_pr_str}",
                           check=True, text=True, shell=True, cwd="/root/sfaira/")
        else:
            subprocess.run(create_pr_str, check=True, text=True, shell=True, cwd="/root/sfaira/")
        subprocess.run(
            ["rm", "-rf", os.path.join(self.path_loader, self.loader_name)], check=False, text=True, shell=False)
        print("[bold green]Your PR was successfully submitted. Feel free to add further comments to it using the URL "
              "in the line above.")



import os

from sfaira.versions.metadata import CelltypeUniverse, OntologyCl, OntologyUberon
from sfaira.unit_tests import DIR_TEMP

"""
CelltypeUniverse
"""


def test_universe_io():
    if not os.path.exists(DIR_TEMP):
        os.mkdir(DIR_TEMP)
    tmp_fn = os.path.join(DIR_TEMP, "universe_temp.csv")
    targets = ["stromal cell", "lymphocyte", "T-helper 1 cell", "T-helper 17 cell"]
    leaves_target = ["stromal cell", "T-helper 1 cell", "T-helper 17 cell"]
    cl = OntologyCl(branch="v2021-02-01")
    uberon = OntologyUberon(branch="2019-11-22")
    cu = CelltypeUniverse(cl=cl, uberon=uberon)
    cu.write_target_universe(fn=tmp_fn, x=targets)
    cu.load_target_universe(fn=tmp_fn)
    os.remove(tmp_fn)
    leaves = cu.onto_cl.convert_to_name(cu.onto_cl.leaves)
    assert set(leaves) == set(leaves_target), (leaves, leaves_target)



import dask.array
import pandas as pd
import numpy as np
import scipy.sparse
import sparse
import torch
from typing import Union


class SfairaDataset(torch.utils.data.Dataset):

    def __init__(
            self,
            map_fn,
            x: Union[np.ndarray, scipy.sparse.spmatrix, dask.array.Array],
            obs: pd.DataFrame,
            **kwargs):
        """

        :param map_fn:
        :param obs:
        :param x:
        :param dask_to_memory: Whether to convert X to a sparse matrix in memory if it is a dask array.
            This implies that x is also moved from disk to memory if it is currently a lazy dask representation of
            a zarr array.
        :param kwargs:
        """
        super(SfairaDataset, self).__init__()
        self.map_fn = map_fn
        self.obs = obs
        if isinstance(x, scipy.sparse.spmatrix) or isinstance(x, sparse.spmatrix):
            self._using_sparse = True
        else:
            self._using_sparse = False
        self.x = x
        assert self.x.shape[0] == self.obs.shape[0], (self.x.shape, self.obs.shape)
        self._len = self.x.shape[0]

    def __len__(self):
        return self._len

    def __getitem__(self, idx):
        if torch.is_tensor(idx):
            idx = idx.tolist()
        if isinstance(idx, int):
            idx = [idx]
        x = self.x[idx, :]
        if self._using_sparse:
            x = x.toarray()
        obs = self.obs.iloc[idx, :]
        xy = self.map_fn(x, obs)
        # Flatten batch dim for torch.Dataset [not necessary for IteratableDataset]
        # TODO this might be inefficient, might need different solution.
        if xy[0][0].shape[0] == 1 and len(xy[0][0].shape) >= 2:
            xy = tuple(tuple(zz.squeeze(axis=0) for zz in z) for z in xy)
        xy = tuple(tuple(torch.from_numpy(zz) for zz in z) for z in xy)
        return xy


class SfairaIterableDataset(torch.utils.data.IterableDataset):

    def __init__(self, iterator_fun, **kwargs):
        super(SfairaIterableDataset, self).__init__(**kwargs)
        self.iterator_fun = iterator_fun

    def __iter__(self):
        return self.iterator_fun()



import abc
import anndata
import dask.dataframe
import numpy as np
import os
import pandas as pd
import sys
from typing import Dict, List, Union


class StoreBase(abc.ABC):
    """
    Base class for store API for attribute typing and shared methods.
    """

    @property
    @abc.abstractmethod
    def adata_by_key(self) -> Dict[str, anndata.AnnData]:
        pass

    @property
    @abc.abstractmethod
    def data_by_key(self):
        pass

    @property
    @abc.abstractmethod
    def indices(self) -> Dict[str, np.ndarray]:
        pass

    @property
    @abc.abstractmethod
    def genome_container(self):
        pass

    @property
    @abc.abstractmethod
    def n_obs(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def n_vars(self):
        pass

    @property
    @abc.abstractmethod
    def obs_by_key(self) -> Dict[str, Union[pd.DataFrame, dask.dataframe.DataFrame]]:
        pass

    @property
    @abc.abstractmethod
    def var_names(self):
        pass

    @property
    @abc.abstractmethod
    def shape(self):
        pass

    @abc.abstractmethod
    def subset(self, attr_key, values: Union[str, List[str], None],
               excluded_values: Union[str, List[str], None], verbose: int):
        pass

    @abc.abstractmethod
    def write_config(self, fn: Union[str, os.PathLike]):
        pass

    @abc.abstractmethod
    def load_config(self, fn: Union[str, os.PathLike]):
        pass

    @abc.abstractmethod
    def checkout(
            self,
            idx: Union[np.ndarray, None],
            batch_size: int,
            obs_keys: List[str],
            return_dense: bool,
            randomized_batch_access: bool,
            random_access: bool,
            **kwargs
    ) -> iter:
        pass

    @property
    def adata_memory_footprint(self) -> Dict[str, float]:
        """
        Memory foot-print of data set k in MB.
        """
        return dict([(k, sys.getsizeof(v) / np.power(1024, 2)) for k, v in self.adata_by_key.items()])



import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "private,GSE115469.csv.gz"
        self.download_url_meta = "private,GSE115469_labels.txt"

        self.assay_sc = "10x 3' v2"
        self.author = "MacParland"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41467-018-06318-7"
        self.layer_counts = "X"
        self.organ = "caudate lobe of liver"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2018

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "celltype"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE115469.csv.gz"),
        os.path.join(data_dir, "GSE115469_labels.txt")
    ]
    adata = anndata.read_csv(fn[0]).T
    celltype_df = pd.read_csv(fn[1], sep="\t").set_index("CellName")
    adata.obs["celltype"] = [str(celltype_df.loc[i]["Cluster#"]) for i in adata.obs.index]

    return adata



import os

import pandas as pd
import scanpy as sc


def load(data_dir, sample_fn, **kwargs):
    adata = sc.read_10x_mtx(data_dir, prefix="GSE190604_")
    fn_meta = os.path.join(data_dir, "GSE190604_cellranger-guidecalls-aggregated-unfiltered.txt.gz")
    tab_meta = pd.read_csv(fn_meta, compression="gzip", sep="\t")
    tab_meta.index = tab_meta["cell_barcode"].values
    del tab_meta["cell_barcode"]
    adata.obs = pd.concat([adata.obs, tab_meta], axis=1)
    return adata



import numpy as np
import pandas as pd
from typing import List, Union

from sfaira.versions.metadata import CelltypeUniverse
from sfaira.consts import OCS
from sfaira.versions.topologies import TopologyContainer, TOPOLOGIES


class ModelZoo:

    """
    Model zoo class.
    """

    _model_id: Union[str, None]
    available_model_ids: Union[list, None]
    celltypes: Union[CelltypeUniverse, None]
    topology_container: Union[None, TopologyContainer]
    zoo: Union[dict, None]

    TOPOLOGIES = TOPOLOGIES
    TOPOLOGY_CONTAINER_CLASS = TopologyContainer

    def __init__(
            self,
            model_lookuptable: Union[None, pd.DataFrame] = None,
            model_class: Union[str, None] = None,
    ):
        """
        :param model_lookuptable: model_lookuptable.
        :param model_class: Model class to subset to.
        """
        self._ontology_container_sfaira = OCS
        self._model_id = None
        self.topology_container = None

        if model_lookuptable is not None:  # check if models in repository
            self._load_model_ids(model_ids=model_lookuptable['model_id'].values, model_class=model_class)
            self._construct_zoo_from_model_ids()
        else:
            self.zoo = None
            self.available_model_ids = None

    def _load_model_ids(
            self,
            model_ids,
            model_class: Union[str, None] = None,
    ):
        """
        Load model ids based on models available in model lookup tables.

        :param model_ids: Table listing all available model_ids.
        :param model_class: Model class to subset to
        """
        self.available_model_ids = [x for x in model_ids if (x.split('_')[0] == model_class or model_class is None)]

    def _construct_zoo_from_model_ids(self):
        """
        Load model zoo based on models available model_ids.
        """
        id_df = pd.DataFrame(
            [i.split('_')[1:3] for i in self.available_model_ids],
            columns=['name', 'organisation']
        )
        orgs = np.unique(id_df['organisation'])
        zoo = dict.fromkeys(orgs)
        for o in orgs:
            id_df_o = id_df[id_df['organisation'] == o]
            name = np.unique(id_df_o['name'])
            zoo[o] = dict.fromkeys(name)
        self.zoo = zoo

    @staticmethod
    def _order_versions(
            versions: List[str]
    ):
        """
        Order list of versions of the form 'vx.y.z' from newest to oldest.

        :param versions: Unordered list of version IDs.
        :return: Ordered list of versions.
        """
        versions.sort(key=lambda s: [int(u) for u in s.split('.')])
        return versions

    def topology(
            self,
            model_type: str,
            organisation: str
    ) -> List[str]:
        """
        Return list of available model topologies that trained by a given organisation, and a given model

        :param model_type: Identifier of model_type to show versions for.
        :param organisation: Identifier of organisation to show versions for.
        :return: List of versions available.
        """
        assert organisation in self.zoo.keys(), "organisation requested was not found in zoo"
        assert model_type in self.zoo[organisation].keys(), \
            "model_type requested was not found in zoo"
        return self.zoo[organisation][model_type]

    def versions(
            self,
            model_type: str,
            organisation: str,
            model_topology: str
    ) -> List[str]:
        """
        Return list of available model versions of a given organisation for a given organism and organ and model.

        :param model_type: Identifier of model_type to show versions for.
        :param organisation: Identifier of organisation to show versions for.
        :param model_topology: Identifier of model_topology to show versions for.
        :return: List of versions available.
        """
        assert organisation in self.zoo.keys(), "organisation requested was not found in zoo"
        assert model_type in self.zoo[organisation].keys(), \
            "model_type requested was not found in zoo"
        assert model_topology in self.zoo[organisation][model_type].keys(), \
            "model_topology requested was not found in zoo"
        return self.zoo[organisation][model_type][model_topology]

    @property
    def model_hyperparameters(self) -> dict:
        assert self.topology_container is not None, "set model_id first"
        return self.topology_container.topology["hyper_parameters"]

    @property
    def celltypes(self):
        assert self.topology_container is not None, "set model_id first"
        return self.topology_container.topology["output"]["targets"]

    @celltypes.setter
    def celltypes(self, x: List):
        assert self.topology_container is not None, "set model_id first"
        self.topology_container.topology["output"]["targets"] = x

    @property
    def model_id(self):
        return self._model_id

    @model_id.setter
    def model_id(self, x: str):
        """
        Set model ID to a manually supplied ID and automatically set topology container.

        :param x: Model ID to set. Format: modelclass_organism-organ-modeltype-topology-version_organisation
        """
        assert self.available_model_ids is None or x in self.available_model_ids,\
            f"{x} not found in available_model_ids, please check available models using ModelZoo.available_model_ids"
        assert len(x.split('_')) == 3, f'model_id {x} is invalid'
        self._model_id = x
        self.topology_container = self.TOPOLOGY_CONTAINER_CLASS(
            topology=self.TOPOLOGIES[self.model_organism][self.model_class][self.model_type][self.model_topology],
            topology_id=self.model_version
        )

    @property
    def model_class(self):
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[0]

    @property
    def model_name(self):
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[1]

    @property
    def model_organism(self):
        # TODO: this relies on theislab model_name formatting
        assert self.model_id is not None, "set model_id first"
        return {
            "homosapiens": "Homo sapiens",
            "musmusculus": "Mus musculus",
            "human": "Homo sapiens",  # necessary for old sfaira model uploads
            "mouse": "Mus musculus",  # necessary for old sfaira model uploads
        }[self.model_id.split('_')[1].split("-")[0]]

    @property
    def model_organ(self):
        # TODO: this relies on theislab model_name formatting
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[1].split("-")[1]

    @property
    def model_type(self):
        # TODO: this relies on theislab model_name formatting
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[1].split("-")[2]

    @property
    def model_topology(self):
        # TODO: this relies on theislab model_name formatting
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[1].split("-")[3]

    @property
    def model_version(self):
        # TODO: this relies on theislab model_name formatting
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[1].split("-")[4]

    @property
    def organisation(self):
        assert self.model_id is not None, "set model_id first"
        return self.model_id.split('_')[2]



import random
from typing import Dict, List, Tuple, Union

import anndata
import dask.array
import numpy as np
import pandas as pd
import scipy.sparse

from sfaira.data.store.batch_schedule import BATCH_SCHEDULE, BatchDesignBase
from sfaira.data.store.carts.base import CartBase
from sfaira.data.store.carts.utils import split_batch


class _ShuffleBuffer:

    def __init__(self, generator: iter, buffer_size: int):
        if buffer_size < 1:
            raise ValueError('buffer_size should be larger than 0')
        self._g = generator
        self._buffer_size = buffer_size

    @staticmethod
    def buffer_replace(buffer, x):
        idx = random.randint(0, len(buffer) - 1)
        val = buffer[idx]
        buffer[idx] = x
        return val

    def generator(self):
        buffer = []
        for x in self._g:
            if len(buffer) == self._buffer_size:
                yield self.buffer_replace(buffer, x)
            else:
                buffer.append(x)
        random.shuffle(buffer)
        while buffer:
            yield buffer.pop()


class CartSingle(CartBase):

    """
    Cart for a DistributedStoreSingleFeatureSpace().
    """

    _obs_idx: Union[np.ndarray, None]
    _batch_schedule_name: str
    batch_size: int
    obs_keys: List[str]
    schedule: BatchDesignBase
    var: pd.DataFrame
    var_idx: Union[None, np.ndarray]

    def __init__(self, obs_idx, obs_keys, var, var_idx=None, batch_schedule="base", batch_size=1, map_fn=None, **kwargs):
        """

        :param batch_schedule: A valid batch schedule name or a class that inherits from BatchDesignBase.

            - "base": sfaira.data.store.batch_schedule.BatchDesignBasic
            - "balanced": sfaira.data.store.batch_schedule.BatchDesignBalanced
            - "blocks": sfaira.data.store.batch_schedule.BatchDesignBlocks
            - "full":  sfaira.data.store.batch_schedule.BatchDesignFull
            - class: batch_schedule needs to be a class (not instance), subclassing BatchDesignBase.

        :param batch_size: Emission batch size. Must be 1 or 0.
        :param map_fn: Map function to apply to output tuple of raw generator. Each draw i from the generator is then:
            `yield map_fn(x[i, var_idx], obs[i, obs_keys])`
        :param obs_idx: np.ndarray: The observations to emit.
        :param obs_keys: .obs columns to return in the generator. These have to be a subset of the columns available
            in self.adata_by_key.
        :param var_idx: The features to emit.
        :parm split_to_obs: Whether to split tensors to observation-wise slices at the emission stage of the generator.
        """
        self._obs_idx = None
        self._batch_schedule_name = batch_schedule
        self.batch_size = batch_size
        self.map_fn = map_fn
        self.obs_keys = obs_keys
        self.var = var
        self.var_idx = var_idx
        if isinstance(batch_schedule, str):
            if batch_schedule == "blocks":
                # Extract group column values from .obs.
                assert "grouping" in kwargs.keys()
                group_vals = self._obs_full[kwargs["grouping"]].values
                kwargs["grouping"] = group_vals
            batch_schedule = BATCH_SCHEDULE[batch_schedule]
        else:
            if not isinstance(batch_schedule, BatchDesignBase):
                raise ValueError("Either supply a name of a valid batch schedule or a class inherting from "
                                 f"BatchDesignBase to the constructor of GeneratorSingle. Found {type(batch_schedule)}.")
        self.schedule = batch_schedule(**kwargs)
        self.obs_idx = obs_idx  # This needs to be set after .schedule.

    @property
    def _obs_full(self):
        """
        Full meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        raise NotImplementedError()

    def _validate_idx(self, idx: Union[np.ndarray, list]) -> np.ndarray:
        """
        Validate global index vector.
        """
        if len(idx) > 0:
            assert np.max(idx) < self.n_obs, f"maximum of supplied index vector {np.max(idx)} exceeds number of " \
                                             f"modelled observations {self.n_obs}"
            assert len(idx) == len(np.unique(idx)), f"repeated indices in idx: {len(idx) - len(np.unique(idx))}"
            if isinstance(idx, np.ndarray):
                assert len(idx.shape) == 1, idx.shape
                assert idx.dtype == np.int
            else:
                assert isinstance(idx, list)
                assert isinstance(idx[0], int) or isinstance(idx[0], np.int)
        idx = np.asarray(idx)
        return idx

    @property
    def _x_full(self):
        """
        Full data matrix (cells x features) that is emitted in batches by .iterator().
        """
        raise NotImplementedError()

    @property
    def adata(self) -> anndata.AnnData:
        """
        Assembles a slice of this cart based on .obs_idx as an anndata instance.
        """
        return anndata.AnnData(X=self.x, obs=self.obs, var=self.var)

    @property
    def n_obs(self) -> int:
        """Total number of observations in cart."""
        raise NotImplementedError()

    @property
    def n_obs_selected(self) -> int:
        """Total number of selected observations in cart."""
        return len(self.obs_idx)

    @property
    def obs_idx(self):
        """
        Integer observation indices to select from cart. These will be emitted if data is queried.
        """
        return self._obs_idx

    @obs_idx.setter
    def obs_idx(self, x):
        if x is None:
            x = np.arange(0, self.n_obs)
        else:
            x = self._validate_idx(x)
            x = np.sort(x)
        # Only reset if they are actually different:
        if self._obs_idx is None or \
                (self._obs_idx is not None and len(x) != len(self._obs_idx)) or \
                np.any(x != self._obs_idx):
            self._obs_idx = x
            self.schedule.idx = x


class CartAnndata(CartSingle):

    """
    Cart for a DistributedStoreAnndata().
    """

    adata_dict: Dict[str, anndata._core.views.ArrayView]
    return_dense: bool
    single_object: bool

    def __init__(self, adata_dict, return_dense=False, **kwargs):
        self.return_dense = return_dense
        self.single_object = len(adata_dict.keys()) == 1
        self.adata_dict = adata_dict
        super(CartAnndata, self).__init__(**kwargs)

    @property
    def _obs_full(self):
        """
        Full meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        return pd.concat([
            v.obs[self.obs_keys]
            for v in self.adata_dict.values()
        ], axis=0, join="inner", ignore_index=True, copy=False)

    @property
    def _x_full(self):
        """
        Full data matrix (cells x features) that is emitted in batches by .iterator().
        """
        x = scipy.sparse.vstack([
            self._parse_array(v.X, return_dense=False)
            for v in self.adata_dict.values()
        ])
        if self.return_dense and isinstance(x, scipy.sparse.spmatrix):
            x = x.todense()
        return x

    def _iterator(self, repeat: int):
        """
        Iterator over data matrix and meta data table, yields batches of data points.
        """
        keep_repeating = True
        num_repetitions = 0

        # Speed up access to single object by skipping index overlap operations:
        while keep_repeating:
            batches = self.schedule.design
            for idx_i in batches:
                if len(idx_i) > 0:
                    # Match adata objects that overlap to batch:
                    idx_i_dict = self._obs_idx_dict_query(idx=idx_i)
                    if self.batch_size == 1:
                        # Emit each data set separately and avoid concatenation into larger chunks for emission.
                        for k, v in idx_i_dict.items():
                            # I) Prepare data matrix.
                            x = self.adata_dict[k].X[v, :]
                            x = self._parse_array(x=x, return_dense=self.return_dense)
                            # Prepare .obs.
                            obs = self.adata_dict[k].obs[self.obs_keys].iloc[v, :]
                            data_tuple = self.map_fn(x, obs)
                            for data_tuple_i in split_batch(x=data_tuple):
                                yield data_tuple_i
                    else:
                        # Concatenates slices first before returning. Note that this is likely slower than emitting by
                        # observation in most scenarios.
                        # I) Prepare data matrix.
                        x = [
                            self._parse_array(self.adata_dict[k].X[v, :], return_dense=self.return_dense)
                            for k, v in idx_i_dict.items()
                        ]
                        is_dense = isinstance(x[0], np.ndarray)
                        # Concatenate blocks in observation dimension:
                        if len(x) > 1:
                            if is_dense:
                                x = np.concatenate(x, axis=0)
                            else:
                                x = scipy.sparse.vstack(x)
                        else:
                            x = x[0]
                        # Prepare .obs.
                        obs = pd.concat([
                            self.adata_dict[k].obs[self.obs_keys].iloc[v, :]
                            for k, v in idx_i_dict.items()
                        ], axis=0, join="inner", ignore_index=True, copy=False)
                        data_tuple = self.map_fn(x, obs)
                        yield data_tuple

            num_repetitions += 1
            keep_repeating = (num_repetitions < repeat) or (repeat <= 0)

    def iterator(self, repeat: int = 1, shuffle_buffer: int = 0):
        if shuffle_buffer > 2 and self.batch_size == 1:
            return _ShuffleBuffer(self._iterator(repeat=repeat), shuffle_buffer).generator()
        else:
            return self._iterator(repeat=repeat)

    def move_to_memory(self):
        """
        No action.
        """
        pass

    @property
    def n_obs(self) -> int:
        """Total number of observations in cart."""
        return int(np.sum([v.n_obs for v in self.adata_dict.values()]))

    @property
    def n_var(self) -> int:
        """Total number of features defined for return in cart."""
        if self.var_idx is None:
            return self.adata_dict[list(self.adata_dict.keys())[0]].n_var
        else:
            return len(self.var_idx)

    @property
    def obs(self):
        """
        Selected meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        idx_dict = self._obs_idx_dict_query(idx=self.obs_idx)
        return pd.concat([
            self.adata_dict[k].obs[self.obs_keys].iloc[v, :]
            for k, v in idx_dict.items()
        ], axis=0, join="inner", ignore_index=True, copy=False)

    @property
    def x(self):
        """
        Selected data matrix (cells x features) that is emitted in batches by .iterator().
        """
        # Assumes that .X are scipy.sparse.csr or coo
        idx_dict = self._obs_idx_dict_query(idx=self.obs_idx)
        x = scipy.sparse.vstack([
            self._parse_array(self.adata_dict[k].X[v, :], return_dense=False)
            for k, v in idx_dict.items()
        ])
        if self.return_dense and isinstance(x, scipy.sparse.spmatrix):
            x = x.todense()
        return x

    # Methods that are specific to this child class:

    @property
    def _idx_dict_global(self) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
        """
        Indexing helper for selecting cells across a dictionary of anndata instances.

        Increasing indices across data sets which can be concatenated into a single index vector with unique entries
        for cells.

        E.g.: For two data sets of 10 cells each, the return value would be

            {A: ([0..9], [0..9]), B: ([0..9], [10..19])}.
        """
        counter = 0
        indices = {}
        for k, v in self.adata_dict.items():
            indices[k] = (np.arange(0, v.n_obs), np.arange(counter, counter + v.n_obs))
            counter += len(v)
        return indices

    def _obs_idx_dict_query(self, idx: np.ndarray) -> Dict[str, np.ndarray]:
        """
        Dictionary of indices of selected observations by data set key.

        :param idx: Array (global indexing) of selected observations.
        """
        if self.single_object:
            idx_dict = dict([(k, np.sort(idx)) for k in self.adata_dict.keys()])
        else:
            idx_set = set(idx)
            # Return data set-wise index if global index is in target set.
            idx_dict = dict([
                (k, np.sort([x1 for x1, x2 in zip(v1, v2) if x2 in idx_set]))
                for k, (v1, v2) in self._idx_dict_global.items()
            ])
            # Only retain non-empty.
            idx_dict = dict([(k, v) for k, v in idx_dict.items() if len(v) > 0])
        return idx_dict

    def _parse_array(self, x, return_dense):
        # Move from ArrayView to numpy if backed and dense:
        if (
                isinstance(x, anndata._core.views.ArrayView) or
                isinstance(x, anndata._core.views.SparseCSRView) or
                isinstance(x, anndata._core.views.SparseCSCView)
        ):
            x = x.toarray()
        if return_dense:
            x = np.asarray(x.todense()) if isinstance(x, scipy.sparse.spmatrix) else x
        elif isinstance(x, np.ndarray):
            x = scipy.sparse.csr_matrix(x) if isinstance(x, np.ndarray) else x
        if self.var_idx is not None:
            x = x[:, self.var_idx]
        return x


class CartDask(CartSingle):

    """
    Cart for a DistributedStoreDao().
    """

    _x: dask.array
    _obs: pd.DataFrame

    def __init__(self, x, obs, obs_keys, **kwargs):
        self._x = x
        self._obs = obs[obs_keys]
        # Redefine index so that .loc indexing can be used instead of .iloc indexing:
        self._obs.index = np.arange(0, obs.shape[0])
        super(CartDask, self).__init__(obs_keys=obs_keys, **kwargs)

    @property
    def _obs_full(self):
        """
        Full meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        return self._obs

    @property
    def _x_full(self):
        """
        Full data matrix (cells x features) that is emitted in batches by .iterator().
        """
        return self._x

    def _iterator(self, repeat: int):
        """
        Iterator over data matrix and meta data table, yields batches of data points.
        """
        # Can all data sets corresponding to one organism as a single array because they share the second dimension
        # and dask keeps expression data and obs out of memory.
        self.schedule.batchsplits = self._x.chunks[0]

        keep_repeating = True
        num_repetitions = 0

        while keep_repeating:
            for batch_idxs in self.schedule.design:
                if len(batch_idxs) > 0:
                    x_i = self._x[batch_idxs, :]
                    if self.var_idx is not None:
                        x_i = x_i[:, self.var_idx]
                    obs_i = self._obs.iloc[batch_idxs, :]
                    data_tuple = self.map_fn(x_i, obs_i)
                    if self.batch_size == 1:
                        for data_tuple_i in split_batch(x=data_tuple):
                            yield data_tuple_i
                    else:
                        yield data_tuple

            num_repetitions += 1
            keep_repeating = (num_repetitions < repeat) or (repeat <= 0)

    def iterator(self, repeat: int = 1, shuffle_buffer: int = 0):
        """
        Iterator over data matrix and meta data table, yields batches of data points.
        """
        if shuffle_buffer > 2 and self.batch_size == 1:
            return _ShuffleBuffer(self._iterator(repeat=repeat), shuffle_buffer).generator()
        else:
            return self._iterator(repeat=repeat)

    def move_to_memory(self):
        """
        Persist underlying dask array into memory in sparse.CSR format.
        """
        self._x = self._x.map_blocks(scipy.sparse.csr_matrix).persist()

    @property
    def n_obs(self) -> int:
        """Total number of observations in cart."""
        return self._x.shape[0]

    @property
    def n_var(self) -> int:
        """Total number of features defined for return in cart."""
        if self.var_idx is None:
            return self._x.shape[1]
        else:
            return len(self.var_idx)

    @property
    def obs(self):
        """
        Selected meta data matrix (cells x meta data) that is emitted in batches by .iterator().
        """
        return self._obs.iloc[self.schedule.idx, :]

    @property
    def x(self):
        """
        Selected data matrix (cells x features) that is emitted in batches by .iterator() as csr matrix.

        Note: to obtain the dask array instead of a csr matrix, use `self.x_dask`
        """
        if self.var_idx is None:
            return self._x[self.schedule.idx, :].compute()
        else:
            return self._x[self.schedule.idx, :][:, self.var_idx].compute()

    # Methods that are specific to this child class:



import abc

import anndata
import numpy as np
import os
import pathlib
from typing import Union

from sfaira.consts.ontologies import DEFAULT_UBERON, DEFAULT_CL
from sfaira.data import load_store
from sfaira.train import TrainModelCelltype, TrainModelEmbedding
from sfaira.ui import ModelZoo
from sfaira.versions.metadata import CelltypeUniverse, OntologyCl, OntologyUberon

from sfaira.unit_tests.tests_by_submodule.estimators import HelperEstimatorBase, TARGETS
from sfaira.unit_tests import DIR_TEMP


def get_cu():
    """
    Get file name of a target universe for loading by trainer.
    """
    # Create temporary cell type universe to give to trainer.
    if not os.path.exists(DIR_TEMP):
        pathlib.Path(DIR_TEMP).mkdir(parents=True, exist_ok=True)
    fn = os.path.join(DIR_TEMP, "universe_temp.csv")
    cl = OntologyCl(branch=DEFAULT_CL)
    uberon = OntologyUberon(branch=DEFAULT_UBERON)
    cu = CelltypeUniverse(cl=cl, uberon=uberon)
    cu.write_target_universe(fn=fn, x=TARGETS)
    del cu
    return fn


class HelperTrainerBase:

    data: Union[anndata.AnnData, load_store]
    trainer: Union[TrainModelCelltype, TrainModelEmbedding]
    dir_temp = DIR_TEMP

    def __init__(self, zoo: ModelZoo):
        self.model_id = zoo.model_id
        self.tc = zoo.topology_container
        self.run_id = self.model_id + "_cv0"

    def load_adata(self, **kwargs):
        """
        This is inherited from estimator test helper.
        """
        pass

    def load_store(self, **kwargs):
        """
        This is inherited from estimator test helper.
        """
        pass

    def load_data(self, data_type):
        """
        Builds training data according to reference used in model definition.

        :param data_type:
        :return:
        """
        np.random.seed(1)
        if data_type == "adata":
            self.load_adata(organism="Homo sapiens", match_to_reference=self.tc.gc.release)
        else:
            self.load_store(organism="Homo sapiens", match_to_reference=self.tc.gc.release)

    def test_init(self, cls, estimator_kwargs: dict = {}, **kwargs):
        if not os.path.exists(self.dir_temp):
            pathlib.Path(self.dir_temp).mkdir(parents=True, exist_ok=True)
        self.load_data(data_type="adata")
        self.trainer = cls(
            data=self.data,
            model_path=os.path.join(self.dir_temp, "model"),
            **kwargs
        )
        self.trainer.zoo.model_id = self.model_id
        self.trainer.init_estim(override_hyperpar={}, **estimator_kwargs)

    def test_save(self):
        if not os.path.exists(self.dir_temp):
            pathlib.Path(self.dir_temp).mkdir(parents=True, exist_ok=True)
        self.trainer.estimator.train(
            epochs=3,
            max_steps_per_epoch=1,
            test_split=0.1,
            validation_split=0.1,
            optimizer="adam",
            lr=0.005,
        )
        self.trainer.save(
            fn=os.path.join(self.dir_temp, self.run_id), model=True, specific=True
        )


class HelperTrainer(HelperEstimatorBase, HelperTrainerBase):
    pass


def test_save_embedding():
    model_id = "embedding_homosapiens-lung-linear-0.1-0.1_mylab"
    zoo = ModelZoo()
    zoo.model_id = model_id
    test_trainer = HelperTrainer(zoo=zoo)
    test_trainer.test_init(cls=TrainModelEmbedding)
    test_trainer.test_save()


def test_save_celltypes():
    tmp_fn = get_cu()
    model_id = "celltype_homosapiens-lung-mlp-0.0.1-0.1_mylab"
    zoo = ModelZoo()
    zoo.model_id = model_id
    test_trainer = HelperTrainer(zoo=zoo)
    test_trainer.test_init(cls=TrainModelCelltype, fn_target_universe=tmp_fn)
    test_trainer.test_save()



import anndata
import os
import scipy.io
import zipfile
import pandas


def load(data_dir, **kwargs):

    cell_line_dict = {
        '409b2': '409B2',
        'H9': 'WA09',
        'Wibj2': 'HPSI0214i-wibj_2',
        'Sc102a1': 'SC102A-1',
        'Kucg2': 'HPSI0214i-kucg_2',
        'Hoik1': 'HPSI0314i-hoik_1',
        'Sojd3': 'HPSI0314i-sojd_3',
    }

    fn = [
        os.path.join(data_dir, "E-MTAB-7552.processed.3.zip"),
        os.path.join(data_dir, "E-MTAB-7552.processed.1.zip"),
        os.path.join(data_dir, "E-MTAB-7552.processed.7.zip")
    ]
    with zipfile.ZipFile(fn[0]) as archive:
        x = scipy.io.mmread(archive.open('human_cell_counts_GRCh38.mtx')).T.tocsr()
    with zipfile.ZipFile(fn[1]) as archive:
        var = pandas.read_csv(archive.open('genes_GRCh38.txt'), sep="\t", index_col=1, names=['ensembl', 'genetype'])
    with zipfile.ZipFile(fn[2]) as archive:
        obs = pandas.read_csv(archive.open('metadata_human_cells.tsv'), sep="\t", index_col=0)
    adata = anndata.AnnData(X=x, var=var, obs=obs)
    adata.obs["Line"] = [cell_line_dict[x] for x in adata.obs["Line"]]

    return adata



from random import randint

import numpy as np

try:
    import tensorflow as tf
except ImportError:
    tf = None
# ToDo: we are using a lot of tf.keras.backend modules below, can we use tf core instead?


class MaskingDense(tf.keras.layers.Layer):
    """ Just copied code from keras Dense layer and added masking and a few other tricks:
        - Direct auto-regressive connections to output
        - Allows a second (non-autoregressive) input that is fully connected to first hidden
        - Either 1 output or 2 outputs (concatenated) that are separately
          auto-regressive wrt to the input
    """

    def __init__(self, units, out_units,
                 hidden_layers=1,
                 dropout_rate=0.0,
                 random_input_order=False,
                 activation='elu',
                 out_activation='linear',
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 out_kernel_initializer='glorot_uniform',
                 out_bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 name=None,
                 batchnorm=False,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(MaskingDense, self).__init__(name=name, **kwargs)

        self.input_sel = None
        self.random_input_order = random_input_order
        self.rate = min(1., max(0., dropout_rate))
        self.kernel_sels = []
        self.units = units
        self.out_units = out_units
        self.hidden_layers = hidden_layers
        self.activation = tf.keras.activations.get(activation)
        self.out_activation = tf.keras.activations.get(out_activation)  # None gives linear activation
        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.out_kernel_initializer = tf.keras.initializers.get(out_kernel_initializer)
        self.out_bias_initializer = tf.keras.initializers.get(out_bias_initializer)
        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.activity_regularizer = tf.keras.regularizers.get(activity_regularizer)
        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)
        self.batchnorm = batchnorm

    def dropout_wrapper(self, inputs, training):
        if 0. < self.rate < 1.:
            def dropped_inputs():
                return tf.keras.backend.dropout(inputs, self.rate, noise_shape=None, seed=None)
            return tf.keras.backend.in_train_phase(dropped_inputs, inputs, training=training)

        return inputs

    def build_layer_weights(
            self,
            input_dim,
            units,
            use_bias=True,
            is_output=False,
            id=''
    ):
        kernel_initializer = (self.kernel_initializer if not is_output
                              else self.out_kernel_initializer)
        bias_initializer = (self.bias_initializer if not is_output
                            else self.out_bias_initializer)

        kernel = self.add_weight(shape=(input_dim, units),
                                 initializer=kernel_initializer,
                                 name='kernel' + id,
                                 regularizer=self.kernel_regularizer,
                                 constraint=self.kernel_constraint)

        if use_bias:
            bias = self.add_weight(shape=(units,),
                                   initializer=bias_initializer,
                                   name='bias' + id,
                                   regularizer=self.bias_regularizer,
                                   constraint=self.bias_constraint)
        else:
            bias = None

        return kernel, bias

    def build_mask(self, shape, prev_sel, is_output):
        if is_output:
            if shape[-1] == len(self.input_sel):
                input_sel = self.input_sel
            else:
                input_sel = self.input_sel * 2
        else:
            # Disallow D-1 because it would violate auto-regressive property
            # Disallow unconnected units by sampling min from previous layer
            input_sel = [randint(np.min(prev_sel), shape[-1] - 2) for i in range(shape[-1])]

        def vals():
            in_len = len(self.input_sel)
            for x in range(shape[-2]):
                for y in range(shape[-1]):
                    if is_output:
                        yield 1 if prev_sel[x] < input_sel[y % in_len] else 0
                    else:
                        yield 1 if prev_sel[x] <= input_sel[y] else 0

        return tf.keras.backend.constant(list(vals()), dtype='float32', shape=shape), input_sel

    def build(self, input_shape):
        if isinstance(input_shape, list):
            if len(input_shape) != 2:
                raise ValueError('Only list only supported for exactly two inputs')
            input_shape, other_input_shape = input_shape

            # Build weights for other (non-autoregressive) vector
            other_shape = (other_input_shape[-1], self.units)
            self.other_kernel, self.other_bias = self.build_layer_weights(*other_shape, id='_h')

        assert len(input_shape) >= 2
        assert self.out_units == input_shape[-1] or self.out_units == 2 * input_shape[-1]

        self.kernels, self.biases = [], []
        self.kernel_masks, self.kernel_sels = [], []
        self.batch_norms = []
        shape = (input_shape[-1], self.units)

        self.input_sel = np.arange(input_shape[-1])
        if self.random_input_order:
            np.random.shuffle(self.input_sel)
        prev_sel = self.input_sel
        for i in range(self.hidden_layers):
            # Hidden layer
            kernel, bias = self.build_layer_weights(*shape, id=str(i))
            self.kernels.append(kernel)
            self.biases.append(bias)

            # Hidden layer mask
            kernel_mask, kernel_sel = self.build_mask(shape, prev_sel, is_output=False)
            self.kernel_masks.append(kernel_mask)
            self.kernel_sels.append(kernel_sel)

            prev_sel = kernel_sel
            shape = (self.units, self.units)
            self.batch_norms.append(tf.keras.layers.BatchNormalization(center=True, scale=True))

        # Direct connection between input/output
        if self.hidden_layers > 0:
            direct_shape = (input_shape[-1], self.out_units)
            self.direct_kernel, _ = self.build_layer_weights(
                *direct_shape,
                use_bias=False,
                is_output=True,
                id='_direct')
            self.direct_kernel_mask, self.direct_sel = self.build_mask(direct_shape, self.input_sel,
                                                                       is_output=True)

        # Output layer
        out_shape = (self.units, self.out_units)
        self.out_kernel, self.out_bias = self.build_layer_weights(
            *out_shape,
            is_output=True,
            id='_out')
        self.out_kernel_mask, self.out_sel = self.build_mask(out_shape, prev_sel, is_output=True)

        self.built = True

    def call(self, inputs, training=None):
        other_input = None
        if isinstance(inputs, list):
            assert len(inputs) == 2
            assert self.hidden_layers > 0, "other input not supported if no hidden layers"
            assert hasattr(self, 'other_kernel')
            inputs, other_input = inputs

        output = inputs

        if other_input is not None:
            other = tf.keras.backend.dot(other_input, self.other_kernel)
            other = tf.keras.backend.bias_add(other, self.other_bias)
            other = self.activation(other)

        # Hidden layer + mask
        for i in range(self.hidden_layers):
            # i=0: input_dim -> masking_dim
            # i>0: masking_dim -> masking_dim
            weight = self.kernels[i] * self.kernel_masks[i]
            output = tf.keras.backend.dot(output, weight)

            # "other" input
            if i == 0 and other_input is not None:
                output = output + other

            output = tf.keras.backend.bias_add(output, self.biases[i])
            output = self.activation(output)
            if self.batchnorm:
                output = self.batch_norms[i](output)
            output = self.dropout_wrapper(output, training)

        # out_act(bias + (V dot M_v)h(x) + (A dot M_a)x + (other dot M_other)other)
        # masking_dim -> input_dim
        output = tf.keras.backend.dot(output, self.out_kernel * self.out_kernel_mask)

        # Direct connection
        if self.hidden_layers > 0:
            # input_dim -> input_dim
            direct = tf.keras.backend.dot(inputs, self.direct_kernel * self.direct_kernel_mask)
            output = output + direct

        output = tf.keras.backend.bias_add(output, self.out_bias)
        output = self.out_activation(output)
        output = self.dropout_wrapper(output, training)

        return output

    def compute_output_shape(self, input_shape):
        if isinstance(input_shape, list):
            input_shape = input_shape[0]
        return (input_shape[0], self.out_units)



import anndata
import os
import tarfile
import pandas as pd
import scipy.sparse

from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "GSM3589406_PP001swap.filtered.matrix.txt.gz",
    "GSM3589407_PP002swap.filtered.matrix.txt.gz",
    "GSM3589408_PP003swap.filtered.matrix.txt.gz",
    "GSM3589409_PP004swap.filtered.matrix.txt.gz",
    "GSM3589410_PP005swap.filtered.matrix.txt.gz",
    "GSM3589411_PP006swap.filtered.matrix.txt.gz",
    "GSM3589412_PP009swap.filtered.matrix.txt.gz",
    "GSM3589413_PP010swap.filtered.matrix.txt.gz",
    "GSM3589414_PP011swap.filtered.matrix.txt.gz",
    "GSM3589415_PP012swap.filtered.matrix.txt.gz",
    "GSM3589416_PP013swap.filtered.matrix.txt.gz",
    "GSM3589417_PP014swap.filtered.matrix.txt.gz",
    "GSM3589418_PP017swap.filtered.matrix.txt.gz",
    "GSM3589419_PP018swap.filtered.matrix.txt.gz",
    "GSM3589420_PP019swap.filtered.matrix.txt.gz",
    "GSM3589421_PP020swap.filtered.matrix.txt.gz",
]

SAMPLE_DICT = {
    "GSM3589406_PP001swap.filtered.matrix.txt.gz": ["lung", "Donor 1", "healthy"],
    "GSM3589407_PP002swap.filtered.matrix.txt.gz": ["lung", "Donor 1", "stimulated"],
    "GSM3589408_PP003swap.filtered.matrix.txt.gz": ["bone marrow", "Donor 1", "healthy"],
    "GSM3589409_PP004swap.filtered.matrix.txt.gz": ["bone marrow", "Donor 1", "stimulated"],
    "GSM3589410_PP005swap.filtered.matrix.txt.gz": ["lymph node", "Donor 1", "healthy"],
    "GSM3589411_PP006swap.filtered.matrix.txt.gz": ["lymph node", "Donor 1", "stimulated"],
    "GSM3589412_PP009swap.filtered.matrix.txt.gz": ["lung", "Donor 2", "healthy"],
    "GSM3589413_PP010swap.filtered.matrix.txt.gz": ["lung", "Donor 2", "stimulated"],
    "GSM3589414_PP011swap.filtered.matrix.txt.gz": ["bone marrow", "Donor 2", "healthy"],
    "GSM3589415_PP012swap.filtered.matrix.txt.gz": ["bone marrow", "Donor 2", "stimulated"],
    "GSM3589416_PP013swap.filtered.matrix.txt.gz": ["lymph node", "Donor 2", "healthy"],
    "GSM3589417_PP014swap.filtered.matrix.txt.gz": ["lymph node", "Donor 2", "stimulated"],
    "GSM3589418_PP017swap.filtered.matrix.txt.gz": ["blood", "Donor A", "stimulated"],
    "GSM3589419_PP018swap.filtered.matrix.txt.gz": ["blood", "Donor A", "healthy"],
    "GSM3589420_PP019swap.filtered.matrix.txt.gz": ["blood", "Donor B", "stimulated"],
    "GSM3589421_PP020swap.filtered.matrix.txt.gz": ["blood", "Donor B", "healthy"],
}


class Dataset(DatasetBase):
    """
    TODO: move state exact to disease
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE126nnn/GSE126030/suppl/GSE126030_RAW.tar"
        self.download_url_meta = [
            "private,donor1.annotation.txt",
            "private,donor2.annotation.txt"
        ]

        self.assay_sc = "10x 3' transcription profiling"
        self.author = "Szabo"
        self.doi_journal = "10.1038/s41467-019-12464-3"
        self.individual = SAMPLE_DICT[self.sample_fn][1]
        self.layer_counts = "X"
        self.organ = SAMPLE_DICT[self.sample_fn][0]
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = SAMPLE_DICT[self.sample_fn][2]
        self.year = 2019

        self.feature_id_var_key = "Accession"
        self.feature_symbol_var_key = "Gene"
        self.feature_type = "rna"

        self.cell_type_obs_key = "cell_ontology_class"

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE126030_RAW.tar"),
        os.path.join(data_dir, "donor1.annotation.txt"),
        os.path.join(data_dir, "donor2.annotation.txt")
    ]
    with tarfile.open(fn[0]) as tar:
        df = pd.read_csv(tar.extractfile(sample_fn), compression="gzip", sep="\t")
    df.index = [i.split(".")[0] for i in df["Accession"]]
    var = pd.concat([df.pop(x) for x in ["Gene", "Accession"]], 1)
    if df.columns[-1].startswith("Un"):
        df.drop(df.columns[len(df.columns) - 1], axis=1, inplace=True)
    adata = anndata.AnnData(df.T)
    adata.var = var
    adata.obs["donor"] = SAMPLE_DICT[sample_fn][1]
    adata.obs.index = sample_fn.split("_")[1].split("s")[0] + "nskept." + adata.obs.index
    adata.obs["cell_ontology_class"] = "unknown"
    df1 = pd.read_csv(fn[1], sep="\t", index_col=0, header=None)
    df2 = pd.read_csv(fn[2], sep="\t", index_col=0, header=None)
    for i in df1.index:
        adata.obs["cell_ontology_class"].loc[i] = df1.loc[i][1]
    for i in df2.index:
        adata.obs["cell_ontology_class"].loc[i] = df2.loc[i][1]
    adata.X = scipy.sparse.csc_matrix(adata.X)

    return adata



from sfaira.models.embedding.base import BasicModelKerasEmbedding
from sfaira.models.embedding.ae import ModelKerasAe, ModelAeVersioned
from sfaira.models.embedding.vae import ModelKerasVae, ModelVaeVersioned
from sfaira.models.embedding.linear import ModelKerasLinear, ModelLinearVersioned
from sfaira.models.embedding.vaeiaf import ModelKerasVaeIAF, ModelVaeIAFVersioned
from sfaira.models.embedding.vaevamp import ModelKerasVaeVamp, ModelVaeVampVersioned



from . import homosapiens
from . import musmusculus
from .class_interface import TopologyContainer

TOPOLOGIES = {
    "Mus musculus": {
        "celltype": {
            "marker": musmusculus.celltype.celltypemarker.CELLTYPEMARKER_TOPOLOGIES,
            "mlp": musmusculus.celltype.celltypemlp.CELLTYPEMLP_TOPOLOGIES
        },
        "embedding": {
            "ae": musmusculus.embedding.ae.AE_TOPOLOGIES,
            "linear": musmusculus.embedding.linear.LINEAR_TOPOLOGIES,
            "nmf": musmusculus.embedding.nmf.NMF_TOPOLOGIES,
            "vae": musmusculus.embedding.vae.VAE_TOPOLOGIES,
            "vaeiaf": musmusculus.embedding.vaeiaf.VAEIAF_TOPOLOGIES,
            "vaevamp": musmusculus.embedding.vaevamp.VAEVAMP_TOPOLOGIES
        }
    },
    "Homo sapiens": {
        "celltype": {
            "marker": homosapiens.celltype.celltypemarker.CELLTYPEMARKER_TOPOLOGIES,
            "mlp": homosapiens.celltype.celltypemlp.CELLTYPEMLP_TOPOLOGIES
        },
        "embedding": {
            "ae": homosapiens.embedding.ae.AE_TOPOLOGIES,
            "linear": homosapiens.embedding.linear.LINEAR_TOPOLOGIES,
            "nmf": homosapiens.embedding.nmf.NMF_TOPOLOGIES,
            "vae": homosapiens.embedding.vae.VAE_TOPOLOGIES,
            "vaeiaf": homosapiens.embedding.vaeiaf.VAEIAF_TOPOLOGIES,
            "vaevamp": homosapiens.embedding.vaevamp.VAEVAMP_TOPOLOGIES
        }
    }
}



import abc
import os
import pickle
from typing import Dict, List, Tuple, Union

import anndata
import dask.array
import dask.dataframe
import numpy as np
import pandas as pd
from sfaira.consts import AdataIdsSfaira, OCS
from sfaira.data.dataloaders.base.utils import is_child, UNS_STRING_META_IN_OBS
from sfaira.data.store.carts.single import CartAnndata, CartDask, CartSingle
from sfaira.data.store.stores.base import StoreBase
from sfaira.versions.genomes.genomes import GenomeContainer, ReactiveFeatureContainer

"""
Distributed stores are array-like classes that sit on groups of on disk representations of anndata instances files.
Depending on the file format of the count matrix on disk, different in memory representations are sensible.
In particular, if .X is saved as zarr array, one can use lazy dask arrays to operate across sets of count matrices,
heavily reducing the complexity of the code required here and often increasing access speed.

This instances sit on groups of data objects that have the same feature space, e.g. are from the same organism.
You can operate on multiple organisms at the same time by using an umbrella class over a set of such instances by using
DistributedStoreMultipleFeatureSpaceBase.

DistributedStoreBase is base class for any file format on disk.
DistributedStoreDao wraps an on-disk representation of anndata instance in the sfaira "dao" format.
DistributedStoreH5ad wraps an on-disk representation of anndata instances as a h5ad file.
DistributedStoreAnndata wraps in-memory anndata instance.

Note that in all cases, you can use standard anndata reading functions to load a single object into memory.
"""


def _process_batch_size(batch_size: int, retrival_batch_size: int) -> Tuple[int, int]:
    if batch_size not in [0, 1]:
        raise ValueError("batch size is only supported other than 0 or 1")
    return batch_size, retrival_batch_size


class StoreSingleFeatureSpace(StoreBase):

    """
    Data set group class tailored to data access requirements common in high-performance computing (HPC).

    This class does not inherit from DatasetGroup because it entirely relies on the cached objects.
    This class is centred around .adata_by_key and .indices.

    .adata_by_key is a dictionary (by id) of backed anndata instances that point to individual h5ads.
    This dictionary is intialised with all h5ads in the store.
    As the store is sub-setted, key-value pairs are deleted from this dictionary.

    .indices have keys that correspond to keys in .adata_by_key and contain index vectors of observations in the anndata
    instances in .adata_by_key which are still kept.
    These index vectors are a form of lazy slicing that does not require data set loading or re-writing.
    As the store is sub-setted, key-value pairs are deleted from this dictionary if no observations from a given key
    match the sub-setting.
    If a subset of observations from a key matches the subsetting operation, the index set in the corresponding value is
    reduced.
    All data retrieval operations work on .indices: Generators run over these indices when retrieving observations for
    example.
    """

    _adata_by_key: Dict[str, anndata.AnnData]
    _indices: Dict[str, np.ndarray]
    _obs_by_key: Union[None, Dict[str, dask.dataframe.DataFrame]]
    data_source: str
    _dataset_weights: Dict[str, float]

    def __init__(self, adata_by_key: Dict[str, anndata.AnnData], indices: Dict[str, np.ndarray],
                 obs_by_key: Union[None, Dict[str, dask.dataframe.DataFrame]] = None, data_source: str = "X"):
        self.adata_by_key = self.__align_categorical_levels(adata_by_key)
        self.indices = indices
        self.obs_by_key = obs_by_key
        self.ontology_container = OCS
        self._genome_container = None
        self._adata_ids_sfaira = AdataIdsSfaira()
        self.data_source = data_source
        self._celltype_universe = None

    @staticmethod
    def __align_categorical_levels(adata_by_key: Dict[str, anndata.AnnData]) -> Dict[str, anndata.AnnData]:
        """
        Align the categorical levels across all datasets.

        :param adata_by_key: Dict[str, anndata.Anndata]
        :return: Dict[str, anndata.Anndata]
        """
        datasets = list(adata_by_key.keys())
        # get list of all categorical columns - using one dataset is enough as they all have the same columns
        categorical_columns: List[str] = []
        for col in adata_by_key[datasets[0]].obs.columns:
            if isinstance(adata_by_key[datasets[0]].obs[col].dtype, pd.api.types.CategoricalDtype):
                categorical_columns.append(col)
        # union categorical levels across datasets for each column
        categories_columns: Dict[str, pd.Index] = {}
        for col in categorical_columns:
            categories_columns[col] = pd.api.types.union_categoricals(
                [pd.Categorical(v.obs[col].cat.categories) for v in adata_by_key.values()]
            ).categories
        # update categorical columns
        for dataset in datasets:
            for col, categories in categories_columns.items():
                adata_by_key[dataset].obs[col] = pd.Categorical(adata_by_key[dataset].obs[col],
                                                                categories=categories)

        return adata_by_key

    @property
    def idx(self) -> np.ndarray:
        """
        Global indices.
        """
        idx_global = np.arange(0, np.sum([len(v) for v in self.indices.values()]))
        return idx_global

    @property
    def organisms_by_key(self) -> Dict[str, str]:
        """
        Data set-wise organism label as dictionary of data set keys.
        """
        ks = self.indices.keys()
        organisms = [self._adata_by_key[k].uns[self._adata_ids_sfaira.organism] for k in ks]
        # Flatten list, assumes that each data set maps to one organism:
        organisms = [x[0] if (isinstance(x, list) or isinstance(x, tuple)) else x for x in organisms]
        return dict(list(zip(ks, organisms)))

    @property
    def organism(self):
        """
        Organism of store.
        """
        organisms = np.unique(list(self.organisms_by_key.values()))
        assert len(organisms) == 1, organisms
        return organisms[0]

    def _validate_feature_space_homogeneity(self) -> List[str]:
        """
        Assert that the data sets which were kept have the same feature names.

        :return: List of feature names in shared feature space or dictionary of list of features.
        """
        reference_k = list(self._adata_by_key.keys())[0]
        var_names = self._adata_by_key[reference_k].var_names.tolist()
        for k in list(self._adata_by_key.keys()):
            assert len(var_names) == len(self._adata_by_key[k].var_names), \
                f"number of features in store differed in object {k} compared to {reference_k}"
            assert np.all(var_names == self._adata_by_key[k].var_names), \
                f"var_names in store were not matched in object {k} compared to {reference_k}"
        return var_names

    @property
    def adata_by_key(self) -> Dict[str, anndata.AnnData]:
        """
        Anndata instance for each selected data set in store, sub-setted by selected cells.
        """
        return self._adata_by_key

    @adata_by_key.setter
    def adata_by_key(self, x: Dict[str, anndata.AnnData]):
        self._adata_by_key = x

    @property
    def data_by_key(self):
        """
        Data matrix for each selected data set in store, sub-setted by selected cells.
        """
        return dict([(k, v.X) for k, v in self.adata_by_key.items()])

    @property
    def indices(self) -> Dict[str, np.ndarray]:
        """
        Indices of observations that are currently exposed in adata of this instance.

        This depends on previous subsetting.
        """
        return self._indices

    @indices.setter
    def indices(self, x: Dict[str, np.ndarray]):
        """
        Setter imposes a few constraints on indices:

            1) checks that keys are contained ._adata_by_key.keys()
            2) checks that indices are contained in size of values of ._adata_by_key
            3) checks that indces are not duplicated
            4) checks that indices are sorted
        """
        for k, v in x.items():
            assert k in self._adata_by_key.keys(), f"did not find key {k}"
            assert np.max(v) < self._adata_by_key[k].n_obs, f"found index for key {k} that exceeded data set size"
            assert len(v) == len(np.unique(v)), f"found duplicated indices for key {k}"
            assert np.all(np.diff(v) >= 0), f"indices not sorted for key {k}"
        self._indices = x

    @property
    def obs_by_key(self) -> Dict[str, Union[pd.DataFrame, dask.dataframe.DataFrame]]:
        if self._obs_by_key is not None:
            # Run sanity checks to validate that this external obs can be used in the context of .adata_by_key:
            assert np.all(list(self._adata_by_key.keys()) == list(self._obs_by_key.keys()))
            assert np.all([self._obs_by_key[k].shape[0] == self._adata_by_key[k].shape[0]
                           for k in self._obs_by_key.keys()])
            return self._obs_by_key
        else:
            return dict([(k, v.obs) for k, v in self.adata_by_key.items()])

    @obs_by_key.setter
    def obs_by_key(self, x: Union[None, Dict[str, dask.dataframe.DataFrame]]):
        if x is not None:
            for k, v in x.items():
                if not (isinstance(v, dask.dataframe.DataFrame) or isinstance(v, pd.DataFrame)):
                    raise ValueError(f"value of entry {k} was not a dask.dataframe.DataFrame but {type(v)}")
        self._obs_by_key = x

    @property
    def genome_container(self) -> Union[GenomeContainer, None]:
        return self._genome_container

    @genome_container.setter
    def genome_container(self, x: Union[GenomeContainer, None]):
        if x is not None:
            var_names = self._validate_feature_space_homogeneity()
            if isinstance(x, ReactiveFeatureContainer):
                # Load stores feature names into container.
                x.symbols = var_names
            else:
                # Validate genome container choice:
                # Make sure that all var names defined in genome container are also contained in loaded data sets.
                assert np.all([y in var_names for y in x.ensembl]), \
                    "did not find variable names from genome container in store"
        self._genome_container = x

    @property
    def dataset_weights(self):
        return self._dataset_weights

    @dataset_weights.setter
    def dataset_weights(self, x: Dict[str, float]):
        assert np.all([k in self.adata_by_key.keys() for k in x.keys()]), "did not recognize some keys"
        assert np.all([k in x.keys() for k in self.indices.keys()]), "some data sets in index were omitted"
        self._dataset_weights = x

    def get_subset_idx(self, attr_key, values: Union[str, List[str], None],
                       excluded_values: Union[str, List[str], None]) -> dict:
        """
        Get indices of subset list of adata objects based on cell-wise properties.

        :param attr_key: Property to subset by. Options:

            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to. Supply either values or excluded_values.
        :param excluded_values: Classes to exclude from match list. Supply either values or excluded_values.
        :return dictionary of files and observation indices by file.
        """
        if not isinstance(values, list):
            values = [values]
        assert (values is None or excluded_values is not None) or (values is not None or excluded_values is None), \
            "supply either values or excluded_values"

        def get_idx(adata, obs, k, v, xv, dataset) -> np.ndarray:
            # Use cell-wise annotation if data set-wide maps are ambiguous:
            # This can happen if the different cell-wise annotations are summarised as a union in .uns.
            read_from_uns = (getattr(self._adata_ids_sfaira, k) in adata.uns.keys() and
                             np.all(adata.uns[getattr(self._adata_ids_sfaira, k)] != UNS_STRING_META_IN_OBS) and
                             getattr(self._adata_ids_sfaira, k) not in obs.columns)
            read_from_obs = not read_from_uns and getattr(self._adata_ids_sfaira, k) in obs.columns
            if read_from_uns:
                values_found = adata.uns[getattr(self._adata_ids_sfaira, k)]
                if isinstance(values_found, np.ndarray):
                    values_found = values_found.tolist()
                elif not isinstance(values_found, list):
                    values_found = [values_found]
                if len(values_found) > 1:
                    values_found = None  # Go to cell-wise annotation.
                else:
                    # Replicate unique property along cell dimension.
                    values_found = [values_found[0] for _ in range(adata.n_obs)]
            elif read_from_obs:
                values_found = obs[getattr(self._adata_ids_sfaira, k)].to_numpy()
            else:
                values_found = []
                print(f"WARNING: did not find attribute {k} in data set {dataset}")

            try:
                ontology = getattr(self.ontology_container, k)
            except AttributeError:
                raise ValueError(f"{k} not a valid property of ontology_container object")

            values_found_unique_matched = []

            unknown_identifiers = [
                self._adata_ids_sfaira.unknown_metadata_identifier,
                self._adata_ids_sfaira.not_a_cell_celltype_identifier
            ]
            if v:
                for unknown_ident in unknown_identifiers:
                    if unknown_ident in v:
                        v.remove(unknown_ident)
                        values_found_unique_matched.append(unknown_ident)
            elif xv:
                for unknown_ident in unknown_identifiers:
                    if unknown_ident in xv:
                        xv.remove(unknown_ident)
                        values_found_unique_matched.append(unknown_ident)

            for x in pd.unique(values_found):
                if x in unknown_identifiers:
                    pass
                elif v is not None and np.any([is_child(query=x, ontology=ontology, ontology_parent=y) for y in v]):
                    values_found_unique_matched.append(x)
                elif xv is not None and np.all([
                    not is_child(query=x, ontology=ontology, ontology_parent=y) for y in xv
                ]):
                    values_found_unique_matched.append(x)

            idx = np.where(np.isin(values_found, values_found_unique_matched))[0]
            return idx

        indices = {}
        for key in self.indices.keys():
            if key not in self.adata_by_key.keys():
                raise ValueError(f"data set {key} queried by indices does not exist in store (.adata_by_key)")
            # Get indices of idx_old to keep:
            adata_k = self.adata_by_key[key]
            obs_k = self.obs_by_key[key]
            idx_old = self.indices[key]
            # Cannot index on view here as indexing on view of views of backed anndata objects is not yet supported.
            idx_subset = get_idx(adata=adata_k, obs=obs_k, k=attr_key, v=values, xv=excluded_values, dataset=key)
            # Keep intersection of old and new hits.
            idx_new = np.intersect1d(idx_old, idx_subset)
            if len(idx_new) > 0:
                indices[key] = np.asarray(idx_new, dtype="int32")

        return indices

    def subset(self, attr_key, values: Union[str, List[str], None] = None,
               excluded_values: Union[str, List[str], None] = None, verbose: int = 1):
        """
        Subset list of adata objects based on cell-wise properties.

        Subsetting is done based on index vectors, the objects remain untouched.

        :param attr_key: Property to subset by. Options:

            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to. Supply either values or excluded_values.
        :param excluded_values: Classes to exclude from match list. Supply either values or excluded_values.
        :param verbose: If >1 print warning message if store is empty after subsetting
        """
        self.indices = self.get_subset_idx(attr_key=attr_key, values=values, excluded_values=excluded_values)
        if self.n_obs == 0 and verbose > 0:
            print(f"WARNING: store is now empty after subsetting {attr_key} for {values}, excluding {excluded_values}.")

    def write_config(self, fn: Union[str, os.PathLike]):
        """
        Writes a config file that describes the current data sub-setting.

        This config file can be loaded later to recreate a sub-setting.
        This config file contains observation-wise subsetting information.

        :param fn: Output file without file type extension.
        """
        with open(fn + '.pickle', 'wb') as f:
            pickle.dump(self.indices, f)

    def load_config(self, fn: Union[str, os.PathLike]):
        """
        Load a config file and recreates a data sub-setting.
        This config file contains observation-wise subsetting information.

        :param fn: Output file without file type extension.
        """
        with open(fn, 'rb') as f:
            self.indices = pickle.load(f)
        # Make sure all declared data sets are found in store:
        for x in self.indices.keys():
            if x not in self.adata_by_key.keys():
                raise ValueError(f"did not find object with name {x} in currently loaded universe")

    @property
    def var_names(self) -> List[str]:
        """
        Feature names of selected genes by organism in store.
        """
        var_names = self._validate_feature_space_homogeneity()
        # Use feature space sub-selection based on assembly if provided, will use full feature space otherwise.
        if self.genome_container is None:
            return var_names
        else:
            return self.genome_container.ensembl

    @property
    def n_vars(self) -> int:
        """
        Number of selected features per organism in store
        """
        var_names = self._validate_feature_space_homogeneity()
        # Use feature space sub-selection based on assembly if provided, will use full feature space otherwise.
        if self.genome_container is None:
            return len(var_names)
        else:
            return self.genome_container.n_var

    @property
    def n_obs(self) -> int:
        """
        Number of observations selected in store.
        """
        return np.sum([len(v) for v in self.indices.values()])

    @property
    def shape(self) -> Tuple[int, int]:
        return self.n_obs, self.n_vars

    def _index_curation_helper(
            self,
            batch_size: int,
            retrival_batch_size: int,
    ) -> Tuple[Union[np.ndarray, None], int, int]:
        """
        Process indices and batch size input for generator production.

        Feature indices are formatted based on previously loaded genome container.

        :param batch_size: Number of observations read from disk in each batched access (generator invocation).
        :return: Tuple:
            - var_idx: Processed feature index vector for generator to access.
            - batch_size: Processed batch size for generator to access. Choose as 0 to keep batch schedule batches.
            - retrival_batch_size: Processed retrieval batch size for generator to access.
        """
        # Make sure that features are ordered in the same way in each object so that generator yields consistent cell
        # vectors.
        var_names = self._validate_feature_space_homogeneity()
        # Use feature space sub-selection based on assembly if provided, will use full feature space otherwise.
        if self.genome_container is not None:
            var_names_target = self.genome_container.ensembl
            if var_names_target is None:
                # Check if genome container does not constrain features.
                var_idx = None
            elif len(var_names_target) == len(var_names) and np.all(var_names_target == var_names):
                # Check if index vector is just full ordered list of indices, in this case, sub-setting is unnecessary.
                var_idx = None
            else:
                # Check if variable names are continuous stretch in reference list, indexing this is much faster.
                # Note: There is about 5 sec to be saved on a call because if len(var_names_target) calls to .index
                #  on a list of length var_names are avoided.
                #  One example in this would save about 5 sec would be selection of protein coding genes from a full
                #  gene space in which protein coding genes grouped together (this is not the case in the standard
                #  assembly).
                idx_first = var_names.index(var_names_target[0])
                idx_last = idx_first + len(var_names_target)
                if idx_last <= len(var_names) and np.all(var_names_target == var_names[idx_first:idx_last]):
                    var_idx = np.arange(idx_first, idx_last)
                else:
                    var_idx = np.sort([var_names.index(x) for x in var_names_target])
        else:
            var_idx = None
        # Select all cells if idx was None:
        batch_size, retrival_batch_size = _process_batch_size(batch_size=batch_size,
                                                              retrival_batch_size=retrival_batch_size)
        return var_idx, batch_size, retrival_batch_size

    @abc.abstractmethod
    def _get_cart(
            self,
            batch_schedule,
            obs_idx: np.ndarray,
            var_idx: Union[np.ndarray, None],
            map_fn,
            obs_keys: List[str],
            **kwargs
    ) -> iter:
        """
        Yields an instance of CartSingle which can emit an iterator over the data defined in the arguments here.

        :param obs_idx: The observations to emit.
        :param var_idx: The features to emit.
        :param map_fn: Map functino to apply to output tuple of raw generator. Each draw i from the generator is then:
            `yield map_fn(x[i, var_idx], obs[i, obs_keys])`
        :param obs_keys: .obs columns to return in the generator. These have to be a subset of the columns available
            in self.adata_by_key.
        :return: GeneratorSingle instance.
        """
        pass

    def checkout(
            self,
            idx: Union[np.ndarray, None] = None,
            batch_size: int = 1,
            retrieval_batch_size: int = 128,
            map_fn=None,
            obs_keys: List[str] = None,
            return_dense: bool = True,
            randomized_batch_access: bool = False,
            random_access: bool = False,
            batch_schedule: str = "base",
            **kwargs
    ) -> CartSingle:
        """
        Yields an instance of a generator class over observations in the contained data sets.

        Multiple such instances can be emitted by a single store class and point to data stored in this store class.
        Effectively, these generators are heavily reduced pointers to the data in an instance of self.
        A common use case is the instantiation of a training data generator and a validation data generator over a data
        subset defined in this class.

        :param idx: Global idx to query from store. These is an array with indices corresponding to a contiuous index
            along all observations in self.adata_by_key, ordered along a hypothetical concatenation along the keys of
            self.adata_by_key. If None, all observations are selected.
        :param batch_size: Number of observations to yield in each access (generator invocation).
        :param retrieval_batch_size: Number of observations read from disk in each batched access (data-backend
            generator invocation).
        :param map_fn: Map functino to apply to output tuple of raw generator. Each draw i from the generator is then:
            `yield map_fn(x[i, var_idx], obs[i, obs_keys])`
        :param obs_keys: .obs columns to return in the generator. These have to be a subset of the columns available
            in self.adata_by_key.
        :param return_dense: Whether to force return count data .X as dense batches. This allows more efficient feature
            indexing if the store is sparse (column indexing on csr matrices is slow).
        :param randomized_batch_access: Whether to randomize batches during reading (in generator). Lifts necessity of
            using a shuffle buffer on generator, however, batch composition stays unchanged over epochs unless there
            is overhangs in retrieval_batch_size in the raw data files, which often happens and results in modest
            changes in batch composition.
            Do not use randomized_batch_access and random_access.
        :param random_access: Whether to fully shuffle observations before batched access takes place. May
            slow down access compared randomized_batch_access and to no randomization.
            Do not use randomized_batch_access and random_access.
        :param batch_schedule: A valid batch schedule name or a class that inherits from BatchDesignBase.

            - "basic": sfaira.data.store.batch_schedule.BatchDesignBasic
            - "balanced": sfaira.data.store.batch_schedule.BatchDesignBalanced
            - "blocks": sfaira.data.store.batch_schedule.BatchDesignBlocks
            - "full":  sfaira.data.store.batch_schedule.BatchDesignFull
            - class: batch_schedule needs to be a class (not instance), subclassing BatchDesignBase.

        :param kwargs: kwargs for idx_generator chosen.
        :return: Generator function which yields batch_size at every invocation.
            The generator returns a tuple of (.X, .obs).
        """
        if obs_keys is None:
            obs_keys = []
        var_idx, batch_size, retrieval_batch_size = self._index_curation_helper(
            batch_size=batch_size, retrival_batch_size=retrieval_batch_size)
        batch_schedule_kwargs = {"randomized_batch_access": randomized_batch_access,
                                 "random_access": random_access,
                                 "retrieval_batch_size": retrieval_batch_size}
        cart = self._get_cart(batch_schedule=batch_schedule, batch_size=batch_size, map_fn=map_fn, obs_idx=idx,
                              obs_keys=obs_keys, return_dense=return_dense, var=self.var, var_idx=var_idx,
                              **batch_schedule_kwargs, **kwargs)
        return cart

    @property
    def var(self) -> Union[pd.DataFrame]:
        if self.genome_container is None:
            var = pd.DataFrame({}, index=self.var_names)
        else:
            var = pd.DataFrame({
                "ensg": self.genome_container.ensembl,
                "symbol": self.genome_container.symbols,
            }, index=self.var_names)
        return var


class StoreAnndata(StoreSingleFeatureSpace):

    in_memory: bool

    def __init__(self, in_memory: bool = True, **kwargs):
        if "indices" not in kwargs.keys():  # Select all cells if no selection performed via indices.
            assert "adata_by_key" in kwargs.keys(), "supply adata_by_key in constructor call to DistributedStoreAnndata"
            kwargs["indices"] = dict([(k, np.arange(0, v.n_obs)) for k, v in kwargs["adata_by_key"].items()])
        super(StoreAnndata, self).__init__(**kwargs)
        self._x_as_dask = False
        self.in_memory = in_memory

    def _get_cart(self, return_dense: bool = False, **kwargs) -> iter:
        return CartAnndata(adata_dict=self._adata_sliced, return_dense=return_dense, **kwargs)

    # Methods that are specific to this child class:

    @property
    def _adata_sliced(self) -> Dict[str, anndata.AnnData]:
        """
        Only exposes the subset and slices of the adata instances contained in ._adata_by_key defined in .indices.
        """
        return dict([(k, self._adata_by_key[k][v, :]) for k, v in self.indices.items()])


class StoreDao(StoreSingleFeatureSpace):

    _dataset_weights: Union[None, Dict[str, float]]
    _x_dask_cache: Union[None, dask.array.Array]
    _x_by_key: Union[None, dask.array.Array]

    def __init__(self, x_by_key, **kwargs):
        super(StoreDao, self).__init__(**kwargs)
        self._x_as_dask = True
        self._x_by_key = x_by_key
        self._x_dask_cache = None

    @property
    def indices(self) -> Dict[str, np.ndarray]:
        return super(StoreDao, self).indices

    @indices.setter
    def indices(self, x: Dict[str, np.ndarray]):
        """
        Extends setter in super class by wiping .X cache.

        Setter imposes a few constraints on indices:

            1) checks that keys are contained ._adata_by_key.keys()
            2) checks that indices are contained in size of values of ._adata_by_key
            3) checks that indces are not duplicated
            4) checks that indices are sorted
        """
        self._x_dask_cache = None
        for k, v in x.items():
            assert k in self._adata_by_key.keys(), f"did not find key {k}"
            assert np.max(v) < self._adata_by_key[k].n_obs, f"found index for key {k} that exceeded data set size"
            assert len(v) == len(np.unique(v)), f"found duplicated indices for key {k}"
            assert np.all(np.diff(v) >= 0), f"indices not sorted for key {k}"
        self._indices = x

    @property
    def data_by_key(self):
        """
        Data matrix for each selected data set in store, sub-setted by selected cells.
        """
        # Accesses _x_by_key rather than _adata_by_key as long as the dask arrays are stored there.
        return dict([(k, self._x_by_key[k][v, :]) for k, v in self.indices.items()])

    def _get_cart(self, **kwargs) -> CartDask:
        return CartDask(x=self._x, obs=self._obs, **kwargs)

    # Methods that are specific to this child class:

    @property
    def _x(self) -> dask.array.Array:
        """
        One dask array of all cells.

        Requires feature dimension to be shared.
        """
        if self._x_dask_cache is None:
            if self.data_source == "X":
                assert np.all([isinstance(self._x_by_key[k], dask.array.Array) for k in self.indices.keys()])
                arrs_to_concat = []
                for k, v in self.indices.items():
                    if self._x_by_key[k].shape[0] == len(v):
                        arrs_to_concat.append(self._x_by_key[k])
                    elif len(v) == 0:
                        continue
                    else:
                        arrs_to_concat.append(self._x_by_key[k][v, :])
                self._x_dask_cache = dask.optimize(dask.array.vstack(arrs_to_concat))[0]
            else:
                raise ValueError(f"Did not recognise data_source={self.data_source}.")
        return self._x_dask_cache

    @property
    def _obs(self) -> pd.DataFrame:
        """
        Assemble .obs table of subset of selected data.

        Resulting index is increasing integers starting with zero.

        :return: .obs data frame.
        """

        return pd.concat([
            self.adata_by_key[k].obs.loc[self.adata_by_key[k].obs.index[v], :]
            for k, v in self.indices.items()
        ], axis=0, join="inner", ignore_index=True, copy=False)



RELEASE_HUMAN = "104"
RELEASE_MOUSE = "104"
ASSEMBLY_HUMAN = f"Homo_sapiens.GRCh38.{RELEASE_HUMAN}"
ASSEMBLY_MOUSE = f"Mus_musculus.GRCm39.{RELEASE_MOUSE}"

CELLTYPES = ["adventitial cell", "endothelial cell", "acinar cell", "pancreatic PP cell", "type B pancreatic cell"]
CL_VERSION = "v2021-08-10"



import anndata
import os

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "private,fetal_liver_alladata_.h5ad"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Popescu"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41586-019-1652-y"
        self.layer_counts = "X"
        self.organ = "liver"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "healthy"
        self.year = 2019

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "cell.labels"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "fetal_liver_alladata_.h5ad")
    adata = anndata.read(fn)

    return adata



"""
All paths used throughout unit testing for temporary files.
"""

import os
import shutil

DIR_BASE_TESTS = os.path.dirname(__file__)
DIR_MODULE_ROOT = os.path.dirname(DIR_BASE_TESTS)
DIR_TEMP = os.path.join(DIR_BASE_TESTS, "temp")

_DIR_DATA_LOADERS = os.path.join(DIR_TEMP, "loaders")
DIR_DATA_LOADERS_CACHE = os.path.join(_DIR_DATA_LOADERS, "cache")
DIR_DATA_LOADERS_STORE_DAO = os.path.join(_DIR_DATA_LOADERS, "store_dao")
DIR_DATA_LOADERS_STORE_H5AD = os.path.join(_DIR_DATA_LOADERS, "store_h5ad")
_DIR_DATA_DATABASES = os.path.join(DIR_TEMP, "databases")
DIR_DATA_DATABASES_CACHE = os.path.join(_DIR_DATA_DATABASES, "cache")
DIR_DATABASE_STORE_DAO = os.path.join(_DIR_DATA_DATABASES, "store_dao")

DIR_SFAIRA_LOADERS = os.path.join(DIR_MODULE_ROOT, "sfaira", "data", "dataloaders", "loaders")


def save_delete(fn):
    assert str(fn).startswith(DIR_TEMP), f"tried to delete outside of temp directory {fn}"
    if os.path.isdir(fn):
        shutil.rmtree(fn)
    else:
        os.remove(fn)



import datetime
from IPython.display import display, display_javascript, display_html
import json
import pydoc
from typing import Union
import uuid

from sfaira.data.dataloaders.base import DatasetGroup, DatasetSuperGroup

from sfaira.data.dataloaders.databases.cellxgene.cellxgene_loader import Dataset
from sfaira.data.dataloaders.databases.cellxgene.rest_helpers import get_collection, get_collections


class DatasetGroupCellxgene(DatasetGroup):

    collection_id: str

    def __init__(
            self,
            collection_id: str = "default",
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            cache_metadata: bool = False,
            verbose: int = 0,
    ):
        self._collection = None
        dataset_ids = [x["id"] for x in get_collection(collection_id=collection_id)['datasets']]
        if len(dataset_ids) == 0:
            print(f"WARNING: Zero data sets retrieved for cellxgene collection {collection_id}.")
        loader_pydoc_path_sfaira = "sfaira.data.dataloaders.databases.cellxgene.cellxgene_loader"
        load_func = pydoc.locate(loader_pydoc_path_sfaira + ".load")
        datasets = [
            Dataset(
                collection_id=collection_id,
                data_path=data_path,
                meta_path=meta_path,
                cache_path=cache_path,
                load_func=load_func,
                sample_fn=x,
                sample_fns=dataset_ids,
                cache_metadata=cache_metadata,
                verbose=verbose,
            )
            for x in dataset_ids
        ]
        keys = [x.id for x in datasets]
        super().__init__(datasets=dict(zip(keys, datasets)), collection_id=collection_id)

    @property
    def collection(self):
        if self._collection is None:
            self._collection = get_collection(collection_id=self.collection_id)
        return self._collection

    def show_summary(self):
        uuid_session = str(uuid.uuid4())
        display_html('<div id="{}" style="height: 600px; width:100%;"></div>'.format(uuid_session), raw=True)
        display_javascript("""
        require(["https://rawgit.com/caldwell/renderjson/master/renderjson.js"], function() {
          document.getElementById('%s').appendChild(renderjson(%s))
        });
        """ % (uuid_session, json.dumps(self.collection)), raw=True)


class DatasetSuperGroupCellxgene(DatasetSuperGroup):

    def __init__(
            self,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            cache_metadata: bool = False,
            verbose: int = 0,
    ):
        self._collections = None
        # Get all collection IDs and instantiate one data set group per collection.
        # Throw warning in collection are not obtained:
        collections = self.collections
        if len(collections) == 0:
            print("WARNING: Zero cellxgene collections retrieved.")
        # Note that the collection itself is not passed to DatasetGroupCellxgene but only the ID string.
        dataset_groups = [
            DatasetGroupCellxgene(
                collection_id=x["id"],
                data_path=data_path,
                meta_path=meta_path,
                cache_path=cache_path,
                cache_metadata=cache_metadata,
                verbose=verbose,
            )
            for x in collections
        ]
        super().__init__(dataset_groups=dataset_groups)

    @property
    def collections(self):
        """
        Yield all collections available from REST API.
        """
        if self._collections is None:
            self._collections = get_collections()
        return self._collections

    def show_summary(self):
        """
        Prints overview of all collections available.
        """
        display("There are " + str(len(self.collections)) + " public collections sorting by newest first:")
        for collection in sorted(self.collections, key=lambda key: key['created_at'], reverse=True):
            display("id: " + collection['id'] + ' created on: ' +
                    datetime.date.fromtimestamp(collection['created_at']).strftime("%m/%d/%y"))



import anndata
import os
import numpy as np

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    """
    TODO add meta data

    .obs columns Age contains entries ['3m', '6m', '7w', '8w', '9w', '10m', '10w', '11w', '12w', '13w', '13y',
       '14w', '15m', '16w', '17w', '24y', '30m', '35y']
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://covid19.cog.sanger.ac.uk/park20.processed.h5ad"
        self.download_url_meta = None

        self.assay_sc = "10x technology"
        self.author = "Park"
        self.disease = "healthy"
        self.doi_journal = "10.1126/science.aay3224"
        self.individual_obs_key = "donor"
        self.layer_processed = "X"
        self.organ = "thymus"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.sex_obs_key = "Gender"
        self.tech_sample_obs_key = "Sample"
        self.year = 2020

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "Anno_level_fig1"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "park20.processed.h5ad")
    adata = anndata.read(fn)
    adata.X = np.expm1(adata.X)
    adata.obs["Gender"] = [{"Male": "male", "Female": "female"}[x] for x in adata.obs["Gender"].values]

    return adata



import anndata
import numpy as np
import pandas
import zipfile
import tarfile
import os

from sfaira.data import DatasetBase

SAMPLE_FNS = [
    "Bladder_dge.txt.gz",
    "BoneMarrow1_dge.txt.gz",
    # "BoneMarrow2_dge.txt.gz",  # not annotated, potentially bad quality
    # "BoneMarrow3_dge.txt.gz",  # not annotated, potentially bad quality
    "BoneMarrowcKit1_dge.txt.gz",
    "BoneMarrowcKit2_dge.txt.gz",
    "BoneMarrowcKit3_dge.txt.gz",
    "Brain1_dge.txt.gz",
    "Brain2_dge.txt.gz",
    # "CJ7.EB14.Ezh2.1_dge.txt.gz",  # TODO: sort out meta data for these
    # "CJ7.EB14.WT.1_dge.txt.gz",  # TODO: sort out meta data for these
    # "CJ7.EB14.WT.2_dge.txt.gz",  # TODO: sort out meta data for these
    # "EB.Ezh2_dge.txt.gz",  # TODO: sort out meta data for these
    # "EB.WT_dge.txt.gz",  # TODO: sort out meta data for these
    # "EmbryonicMesenchymeE14.5_dge.txt.gz",  # TODO: sort out meta data for these
    # "EmbryonicStemCell.CJ7_Deep_dge.txt.gz",  # TODO: sort out meta data for these
    # "EmbryonicStemCells_dge.txt.gz",  # TODO: sort out meta data for these
    "FetalBrain_dge.txt.gz",
    "FetalFemaleGonad_dge.txt.gz",
    "FetalIntestine_dge.txt.gz",
    "FetalKidney1_dge.txt.gz",
    "FetalKidney2_dge.txt.gz",
    "FetalLiverE14.1_dge.txt.gz",
    "FetalLung_dge.txt.gz",
    "FetalMaleGonad_dge.txt.gz",
    "FetalPancreas_dge.txt.gz",
    "FetalStomach_dge.txt.gz",
    # "homosapiens-293T_dge.txt.gz",  # ToDo: sort out meta data for these
    "Kidney1_dge.txt.gz",
    "Kidney2_dge.txt.gz",
    "Liver1_dge.txt.gz",
    "Liver2_dge.txt.gz",
    "Lung1_dge.txt.gz",
    "Lung2_dge.txt.gz",
    "Lung3_dge.txt.gz",
    # "MammaryGland.Involution.CD45.1_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Involution.CD45.2_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Involution1_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Involution2_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Lactation1_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Lactation2_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Pregnancy_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Virgin.CD45.1_dge.txt.gz",  # TODO not annotated?
    # "MammaryGland.Virgin.CD45.2_dge.txt.gz",  # TODO not annotated?
    "MammaryGland.Virgin1_dge.txt.gz",
    "MammaryGland.Virgin2_dge.txt.gz",
    "MammaryGland.Virgin3_dge.txt.gz",
    "MammaryGland.Virgin4_dge.txt.gz",
    # "mES.CJ7_dge.txt.gz",  # TODO: sort out meta data for these
    "MesenchymalStemCells_dge.txt.gz",
    "MesenchymalStemCellsPrimary_dge.txt.gz",
    # "musmusculus-3T3_dge.txt.gz",  # TODO: sort out meta data for these
    "Muscle_dge.txt.gz",
    "NeonatalCalvaria1_dge.txt.gz",
    "NeonatalCalvaria2_dge.txt.gz",
    "NeonatalHeart_dge.txt.gz",
    "NeonatalMuscle1_dge.txt.gz",
    "NeonatalMuscle2_dge.txt.gz",
    # "NeonatalPancreas_dge.txt.zip",  # TODO enable zip file here
    "NeonatalRib1_dge.txt.gz",
    "NeonatalRib2_dge.txt.gz",
    "NeonatalRib3_dge.txt.gz",
    "NeonatalSkin_dge.txt.gz",
    "NeontalBrain1_dge.txt.gz",
    "NeontalBrain2_dge.txt.gz",
    "Ovary1_dge.txt.gz",
    "Ovary2_dge.txt.gz",
    "Pancreas_dge.txt.gz",
    "PeripheralBlood1_dge.txt.gz",
    "PeripheralBlood2_dge.txt.gz",
    "PeripheralBlood3_dge.txt.gz",
    "PeripheralBlood4_dge.txt.gz",
    "PeripheralBlood5_dge.txt.gz",
    "PeripheralBlood6_dge.txt.gz",
    "PlacentaE14.1_dge.txt.gz",
    "PlacentaE14.2_dge.txt.gz",
    "Prostate1_dge.txt.gz",
    "Prostate2_dge.txt.gz",
    "SmallIntestine.CD45_dge.txt.gz",
    "SmallIntestine1_dge.txt.gz",
    "SmallIntestine2_dge.txt.gz",
    "SmallIntestine3_dge.txt.gz",
    "Spleen_dge.txt.gz",
    "Stomach_dge.txt.gz",
    "Testis1_dge.txt.gz",
    "Testis2_dge.txt.gz",
    "Thymus1_dge.txt.gz",
    "Thymus2_dge.txt.gz",
    "TrophoblastStemCells_dge.txt.gz",
    "Uterus1_dge.txt.gz",
    "Uterus2_dge.txt.gz",
]


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        sample_organ_dict = {
            "Bladder_dge.txt.gz": "urinary bladder",
            "BoneMarrow1_dge.txt.gz": "bone marrow",
            "BoneMarrow2_dge.txt.gz": "bone marrow",
            "BoneMarrow3_dge.txt.gz": "bone marrow",
            "BoneMarrowcKit1_dge.txt.gz": "bone marrow",
            "BoneMarrowcKit2_dge.txt.gz": "bone marrow",
            "BoneMarrowcKit3_dge.txt.gz": "bone marrow",
            "Brain1_dge.txt.gz": "brain",
            "Brain2_dge.txt.gz": "brain",
            "CJ7.EB14.Ezh2.1_dge.txt.gz": None,
            "CJ7.EB14.WT.1_dge.txt.gz": None,
            "CJ7.EB14.WT.2_dge.txt.gz": None,
            "EB.Ezh2_dge.txt.gz": None,
            "EB.WT_dge.txt.gz": None,
            "EmbryonicMesenchymeE14.5_dge.txt.gz": "mesenchyme",
            "EmbryonicStemCell.CJ7_Deep_dge.txt.gz": "blastocyst",
            "EmbryonicStemCells_dge.txt.gz": "blastocyst",
            "FetalBrain_dge.txt.gz": "brain",
            "FetalFemaleGonad_dge.txt.gz": "ovary",
            "FetalIntestine_dge.txt.gz": "intestine",
            "FetalKidney1_dge.txt.gz": "kidney",
            "FetalKidney2_dge.txt.gz": "kidney",
            "FetalLiverE14.1_dge.txt.gz": "liver",
            "FetalLung_dge.txt.gz": "lung",
            "FetalMaleGonad_dge.txt.gz": "testis",
            "FetalPancreas_dge.txt.gz": "pancreas",
            "FetalStomach_dge.txt.gz": "stomach",
            "homosapiens-293T_dge.txt.gz": None,
            "Kidney1_dge.txt.gz": "kidney",
            "Kidney2_dge.txt.gz": "kidney",
            "Liver1_dge.txt.gz": "liver",
            "Liver2_dge.txt.gz": "liver",
            "Lung1_dge.txt.gz": "lung",
            "Lung2_dge.txt.gz": "lung",
            "Lung3_dge.txt.gz": "lung",
            "MammaryGland.Involution.CD45.1_dge.txt.gz": "mammary gland",
            "MammaryGland.Involution.CD45.2_dge.txt.gz": "mammary gland",
            "MammaryGland.Involution1_dge.txt.gz": "mammary gland",
            "MammaryGland.Involution2_dge.txt.gz": "mammary gland",
            "MammaryGland.Lactation1_dge.txt.gz": "mammary gland",
            "MammaryGland.Lactation2_dge.txt.gz": "mammary gland",
            "MammaryGland.Pregnancy_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin.CD45.1_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin.CD45.2_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin1_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin2_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin3_dge.txt.gz": "mammary gland",
            "MammaryGland.Virgin4_dge.txt.gz": "mammary gland",
            "mES.CJ7_dge.txt.gz": "blastocyst",
            "MesenchymalStemCells_dge.txt.gz": "mesenchyme",
            "MesenchymalStemCellsPrimary_dge.txt.gz": "mesenchyme",
            "musmusculus-3T3_dge.txt.gz": None,
            "Muscle_dge.txt.gz": "skeletal muscle organ",
            "NeonatalCalvaria1_dge.txt.gz": "vault of skull",
            "NeonatalCalvaria2_dge.txt.gz": "vault of skull",
            "NeonatalHeart_dge.txt.gz": "heart",
            "NeonatalMuscle1_dge.txt.gz": "skeletal muscle organ",
            "NeonatalMuscle2_dge.txt.gz": "skeletal muscle organ",
            "NeonatalPancreas_dge.txt.zip": "pancreas",
            "NeonatalRib1_dge.txt.gz": "rib",
            "NeonatalRib2_dge.txt.gz": "rib",
            "NeonatalRib3_dge.txt.gz": "rib",
            "NeonatalSkin_dge.txt.gz": "skin of body",
            "NeontalBrain1_dge.txt.gz": "brain",
            "NeontalBrain2_dge.txt.gz": "brain",
            "Ovary1_dge.txt.gz": "ovary",
            "Ovary2_dge.txt.gz": "ovary",
            "Pancreas_dge.txt.gz": "pancreas",
            "PeripheralBlood1_dge.txt.gz": "blood",
            "PeripheralBlood2_dge.txt.gz": "blood",
            "PeripheralBlood3_dge.txt.gz": "blood",
            "PeripheralBlood4_dge.txt.gz": "blood",
            "PeripheralBlood5_dge.txt.gz": "blood",
            "PeripheralBlood6_dge.txt.gz": "blood",
            "PlacentaE14.1_dge.txt.gz": "placenta",
            "PlacentaE14.2_dge.txt.gz": "placenta",
            "Prostate1_dge.txt.gz": "prostate gland",
            "Prostate2_dge.txt.gz": "prostate gland",
            "SmallIntestine.CD45_dge.txt.gz": "small intestine",
            "SmallIntestine1_dge.txt.gz": "small intestine",
            "SmallIntestine2_dge.txt.gz": "small intestine",
            "SmallIntestine3_dge.txt.gz": "small intestine",
            "Spleen_dge.txt.gz": "spleen",
            "Stomach_dge.txt.gz": "stomach",
            "Testis1_dge.txt.gz": "testis",
            "Testis2_dge.txt.gz": "testis",
            "Thymus1_dge.txt.gz": "testis",
            "Thymus2_dge.txt.gz": "testis",
            "TrophoblastStemCells_dge.txt.gz": "trophoblast",
            "Uterus1_dge.txt.gz": "uterus",
            "Uterus2_dge.txt.gz": "uterus",
        }
        sample_dev_stage_dict = {
            "Bladder_dge.txt.gz": "adult",
            "BoneMarrow1_dge.txt.gz": "adult",
            "BoneMarrow2_dge.txt.gz": "adult",
            "BoneMarrow3_dge.txt.gz": "adult",
            "BoneMarrowcKit1_dge.txt.gz": "adult",
            "BoneMarrowcKit2_dge.txt.gz": "adult",
            "BoneMarrowcKit3_dge.txt.gz": "adult",
            "Brain1_dge.txt.gz": "adult",
            "Brain2_dge.txt.gz": "adult",
            "CJ7.EB14.Ezh2.1_dge.txt.gz": None,
            "CJ7.EB14.WT.1_dge.txt.gz": None,
            "CJ7.EB14.WT.2_dge.txt.gz": None,
            "EB.Ezh2_dge.txt.gz": None,
            "EB.WT_dge.txt.gz": None,
            "EmbryonicMesenchymeE14.5_dge.txt.gz": "embryonic",
            "EmbryonicStemCell.CJ7_Deep_dge.txt.gz": "embryonic",
            "EmbryonicStemCells_dge.txt.gz": "embryonic",
            "FetalBrain_dge.txt.gz": "fetal",
            "FetalFemaleGonad_dge.txt.gz": "fetal",
            "FetalIntestine_dge.txt.gz": "fetal",
            "FetalKidney1_dge.txt.gz": "fetal",
            "FetalKidney2_dge.txt.gz": "fetal",
            "FetalLiverE14.1_dge.txt.gz": "fetal",
            "FetalLung_dge.txt.gz": "fetal",
            "FetalMaleGonad_dge.txt.gz": "fetal",
            "FetalPancreas_dge.txt.gz": "fetal",
            "FetalStomach_dge.txt.gz": "fetal",
            "homosapiens-293T_dge.txt.gz": None,
            "Kidney1_dge.txt.gz": "adult",
            "Kidney2_dge.txt.gz": "adult",
            "Liver1_dge.txt.gz": "adult",
            "Liver2_dge.txt.gz": "adult",
            "Lung1_dge.txt.gz": "adult",
            "Lung2_dge.txt.gz": "adult",
            "Lung3_dge.txt.gz": "adult",
            "MammaryGland.Involution.CD45.1_dge.txt.gz": "adult",
            "MammaryGland.Involution.CD45.2_dge.txt.gz": "adult",
            "MammaryGland.Involution1_dge.txt.gz": "adult",
            "MammaryGland.Involution2_dge.txt.gz": "adult",
            "MammaryGland.Lactation1_dge.txt.gz": "adult",
            "MammaryGland.Lactation2_dge.txt.gz": "adult",
            "MammaryGland.Pregnancy_dge.txt.gz": "adult",
            "MammaryGland.Virgin.CD45.1_dge.txt.gz": "adult",
            "MammaryGland.Virgin.CD45.2_dge.txt.gz": "adult",
            "MammaryGland.Virgin1_dge.txt.gz": "adult",
            "MammaryGland.Virgin2_dge.txt.gz": "adult",
            "MammaryGland.Virgin3_dge.txt.gz": "adult",
            "MammaryGland.Virgin4_dge.txt.gz": "adult",
            "mES.CJ7_dge.txt.gz": "embryonic",
            "MesenchymalStemCells_dge.txt.gz": "embryonic",
            "MesenchymalStemCellsPrimary_dge.txt.gz": "embryonic",
            "musmusculus-3T3_dge.txt.gz": None,
            "Muscle_dge.txt.gz": "adult",
            "NeonatalCalvaria1_dge.txt.gz": "neonatal",
            "NeonatalCalvaria2_dge.txt.gz": "neonatal",
            "NeonatalHeart_dge.txt.gz": "neonatal",
            "NeonatalMuscle1_dge.txt.gz": "neonatal",
            "NeonatalMuscle2_dge.txt.gz": "neonatal",
            "NeonatalPancreas_dge.txt.zip": "neonatal",
            "NeonatalRib1_dge.txt.gz": "neonatal",
            "NeonatalRib2_dge.txt.gz": "neonatal",
            "NeonatalRib3_dge.txt.gz": "neonatal",
            "NeonatalSkin_dge.txt.gz": "neonatal",
            "NeontalBrain1_dge.txt.gz": "neonatal",
            "NeontalBrain2_dge.txt.gz": "neonatal",
            "Ovary1_dge.txt.gz": "adult",
            "Ovary2_dge.txt.gz": "adult",
            "Pancreas_dge.txt.gz": "adult",
            "PeripheralBlood1_dge.txt.gz": "adult",
            "PeripheralBlood2_dge.txt.gz": "adult",
            "PeripheralBlood3_dge.txt.gz": "adult",
            "PeripheralBlood4_dge.txt.gz": "adult",
            "PeripheralBlood5_dge.txt.gz": "adult",
            "PeripheralBlood6_dge.txt.gz": "adult",
            "PlacentaE14.1_dge.txt.gz": "fetal",
            "PlacentaE14.2_dge.txt.gz": "fetal",
            "Prostate1_dge.txt.gz": "adult",
            "Prostate2_dge.txt.gz": "adult",
            "SmallIntestine.CD45_dge.txt.gz": "adult",
            "SmallIntestine1_dge.txt.gz": "adult",
            "SmallIntestine2_dge.txt.gz": "adult",
            "SmallIntestine3_dge.txt.gz": "adult",
            "Spleen_dge.txt.gz": "adult",
            "Stomach_dge.txt.gz": "adult",
            "Testis1_dge.txt.gz": "adult",
            "Testis2_dge.txt.gz": "adult",
            "Thymus1_dge.txt.gz": "adult",
            "Thymus2_dge.txt.gz": "adult",
            "TrophoblastStemCells_dge.txt.gz": "embryonic",
            "Uterus1_dge.txt.gz": "adult",
            "Uterus2_dge.txt.gz": "adult",
        }

        self.organ = sample_organ_dict[self.sample_fn]

        self.download_url_data = "https://ndownloader.figshare.com/articles/5435866?private_link=865e694ad06d5857db4b"
        self.download_url_meta = None

        self.author = "Han"
        self.dev_stage = sample_dev_stage_dict[self.sample_fn]
        self.disease = "healthy"
        self.doi_journal = "10.1016/j.cell.2018.02.001"
        self.layer_counts = "X"
        self.primary_data = True
        self.organism = "Mus musculus"
        self.assay_sc = "microwell-seq"
        self.year = 2018
        self.sample_source = "primary_tissue"

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        # Only adult and neonatal samples are annotated:
        self.cell_type_obs_key = "Annotation" \
            if sample_dev_stage_dict[self.sample_fn] in ["adult", "neonatal"] and \
            self.sample_fn not in [
                "NeontalBrain1_dge.txt.gz",
                "NeontalBrain2_dge.txt.gz",
                "SmallIntestine.CD45_dge.txt.gz",
                "Thymus2_dge.txt.gz",
            ] else None

        self.set_dataset_id(idx=1)


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, '5435866.zip')
    with zipfile.ZipFile(fn) as archive:
        celltypes = pandas.read_csv(archive.open('MCA_CellAssignments.csv'), index_col=1)
        celltypes = celltypes.drop(["Unnamed: 0"], axis=1)

        with tarfile.open(fileobj=archive.open('MCA_500more_dge.tar.gz')) as tar:
            data = pandas.read_csv(tar.extractfile(f'500more_dge/{sample_fn}'),
                                   compression="gzip",
                                   sep=" ",
                                   header=0
                                   )

    adata = anndata.AnnData(data.T)
    annotated_cells = np.array([x in celltypes.index for x in adata.obs_names])
    # Add annotation if available for this data set:
    if np.sum(annotated_cells) > 0:
        # Subset to annotated cells if any are annotated:
        adata = adata[annotated_cells].copy()
        # Clean nans in data frame to avoid issues with cell type annotation:
        celltypes["Annotation"] = [
            x if x not in [np.nan, "nan"] else "unknown"
            for x in celltypes["Annotation"].values
        ]
        adata.obs = celltypes.loc[adata.obs_names, :]

    return adata



import anndata
import dask.dataframe
import numpy as np
import os
import pandas as pd
import pickle
from typing import Dict, List, Tuple, Union

from sfaira.consts import AdataIdsSfaira
from sfaira.data.store.stores.base import StoreBase
from sfaira.data.store.stores.single import StoreSingleFeatureSpace, \
    StoreDao, StoreAnndata
from sfaira.data.store.carts.multi import CartMulti
from sfaira.data.store.io.io_dao import read_dao
from sfaira.versions.genomes.genomes import GenomeContainer


class StoreMultipleFeatureSpaceBase(StoreBase):

    """
    Umbrella class for a dictionary over multiple instances DistributedStoreSingleFeatureSpace.

    Allows for operations on data sets that are defined in different feature spaces.
    """

    _adata_ids_sfaira: AdataIdsSfaira
    _stores: Dict[str, StoreSingleFeatureSpace]

    def __init__(self, stores: Dict[str, StoreSingleFeatureSpace]):
        self._stores = stores

    @property
    def stores(self) -> Dict[str, StoreSingleFeatureSpace]:
        """
        Only expose stores that contain observations.
        """
        return dict([(k, v) for k, v in self._stores.items() if v.n_obs > 0])

    @stores.setter
    def stores(self, x: Dict[str, StoreSingleFeatureSpace]):
        raise NotImplementedError("cannot set this attribute, it s defined in constructor")

    @property
    def genome_container(self) -> Dict[str, Union[GenomeContainer, None]]:
        return dict([(k, v.genome_container) for k, v in self._stores.items()])

    @genome_container.setter
    def genome_container(self, x: Union[GenomeContainer, Dict[str, GenomeContainer]]):
        if isinstance(x, GenomeContainer):
            # Transform into dictionary first.
            organisms = [k for k, v in self.stores.items()]
            if isinstance(organisms, list) and len(organisms) == 0:
                raise Warning("found empty organism lists in genome_container.setter")
            if len(organisms) > 1:
                raise ValueError(f"Gave a single GenomeContainer for a store instance that has mulitiple organism: "
                                 f"{organisms}, either further subset the store or give a dictionary of "
                                 f"GenomeContainers")
            else:
                x = {organisms[0]: x}
        for k, v in x.items():
            self.stores[k].genome_container = v

    @property
    def indices(self) -> Dict[str, np.ndarray]:
        """
        Dictionary of indices of selected observations contained in all stores.
        """
        return dict([(kk, vv) for k, v in self.stores.items() for kk, vv in v.indices.items()])

    @property
    def adata_by_key(self) -> Dict[str, anndata.AnnData]:
        """
        Dictionary of all anndata instances for each selected data set in store, sub-setted by selected cells, for each
        stores.
        """
        return dict([(kk, vv) for k, v in self.stores.items() for kk, vv in v.adata_by_key.items()])

    @property
    def data_by_key(self):
        """
        Data matrix for each selected data set in store, sub-setted by selected cells.
        """
        return dict([(kk, vv) for k, v in self.stores.items() for kk, vv in v.data_by_key.items()])

    @property
    def obs_by_key(self) -> Dict[str, Union[pd.DataFrame, dask.dataframe.DataFrame]]:
        """
        Dictionary of all anndata instances for each selected data set in store, sub-setted by selected cells, for each
        stores.
        """
        return dict([(k, v.obs) for k, v in self.adata_by_key.items()])

    @property
    def var_names(self) -> Dict[str, List[str]]:
        """
        Dictionary of variable names by store.
        """
        return dict([(k, v.var_names) for k, v in self.stores.items()])

    @property
    def n_vars(self) -> Dict[str, int]:
        """
        Dictionary of number of features by store.
        """
        return dict([(k, v.n_vars) for k, v in self.stores.items()])

    @property
    def n_obs(self) -> int:
        """
        Dictionary of number of observations across stores.
        """
        return np.asarray(np.sum([v.n_obs for v in self.stores.values()]), dtype="int32")

    @property
    def n_obs_dict(self) -> Dict[str, int]:
        """
        Dictionary of number of observations by store.
        """
        return dict([(k, v.n_obs) for k, v in self.stores.items()])

    @property
    def shape(self) -> Dict[str, Tuple[int, int]]:
        """
        Dictionary of full selected data matrix shape by store.
        """
        return dict([(k, v.shape) for k, v in self.stores.items()])

    def subset(self, attr_key, values: Union[str, List[str], None] = None,
               excluded_values: Union[str, List[str], None] = None, verbose: int = 1):
        """
        Subset list of adata objects based on cell-wise properties.

        Subsetting is done based on index vectors, the objects remain untouched.

        :param attr_key: Property to subset by. Options:

            - "assay_differentiation" points to self.assay_differentiation_obs_key
            - "assay_sc" points to self.assay_sc_obs_key
            - "assay_type_differentiation" points to self.assay_type_differentiation_obs_key
            - "cell_line" points to self.cell_line
            - "cell_type" points to self.cell_type_obs_key
            - "developmental_stage" points to self.developmental_stage_obs_key
            - "ethnicity" points to self.ethnicity_obs_key
            - "organ" points to self.organ_obs_key
            - "organism" points to self.organism_obs_key
            - "sample_source" points to self.sample_source_obs_key
            - "sex" points to self.sex_obs_key
            - "state_exact" points to self.state_exact_obs_key
        :param values: Classes to overlap to. Supply either values or excluded_values.
        :param excluded_values: Classes to exclude from match list. Supply either values or excluded_values.
        """
        for k in self.stores.keys():
            self.stores[k].subset(attr_key=attr_key, values=values, excluded_values=excluded_values, verbose=0)
        if self.n_obs == 0 and verbose > 0:
            print("WARNING: multi store is now empty.")

    def write_config(self, fn: Union[str, os.PathLike]):
        """
        Writes a config file that describes the current data sub-setting.

        This config file can be loaded later to recreate a sub-setting.
        This config file contains observation-wise subsetting information.

        :param fn: Output file without file type extension.
        """
        indices = {}
        for v in self.stores.values():
            indices.update(v.indices)
        with open(fn + '.pickle', 'wb') as f:
            pickle.dump(indices, f)

    def load_config(self, fn: Union[str, os.PathLike]):
        """
        Load a config file and recreates a data sub-setting.
        This config file contains observation-wise subsetting information.

        :param fn: Output file without file type extension.
        """
        with open(fn, 'rb') as f:
            indices = pickle.load(f)
        # Distribute indices to corresponding stores by matched keys.
        keys_found = []
        for k, v in self.stores.items():
            indices_k = {}
            for kk, vv in indices.items():
                if kk in v.adata_by_key.keys():
                    indices_k[kk] = vv
                    keys_found.append(kk)
            self.stores[k].indices = indices_k
        # Make sure all declared data were assigned to stores:
        keys_not_found = list(set(list(indices.keys())).difference(set(keys_found)))
        if len(keys_not_found) > 0:
            raise ValueError(f"did not find object(s) with name(s) in store: {keys_not_found}")

    def checkout(
            self,
            idx: Union[Dict[str, Union[np.ndarray, None]], None] = None,
            intercalated: bool = True,
            **kwargs
    ) -> CartMulti:
        """
        Carts per store.

        See also DistributedStore*.checkout().

        :param idx:
        :param intercalated: Whether to do sequential or intercalated emission.
        :param kwargs: See parameters of DistributedStore*.generator().
        :return: Generator function which yields batch_size at every invocation.
            The generator returns a tuple of (.X, .obs).
        """
        if idx is None:
            idx = dict([(k, None) for k in self.stores.keys()])
        for k in self.stores.keys():
            assert k in idx.keys(), (idx.keys(), self.stores.keys())
        carts = dict([(k, v.checkout(idx=idx[k], **kwargs)) for k, v in self.stores.items()])
        return CartMulti(carts=carts, intercalated=intercalated)


class StoresAnndata(StoreMultipleFeatureSpaceBase):

    def __init__(self, adatas: Union[anndata.AnnData, List[anndata.AnnData], Tuple[anndata.AnnData]]):
        # Collect all data loaders from files in directory:
        self._adata_ids_sfaira = AdataIdsSfaira()
        adata_by_key = {}
        indices = {}
        if isinstance(adatas, anndata.AnnData):
            adatas = [adatas]
        for i, adata in enumerate(adatas):
            # Check if adata has a unique ID, if not, add one:
            if self._adata_ids_sfaira.id not in adata.uns.keys():
                adata.uns[self._adata_ids_sfaira.id] = f"adata_{i}"
            if self._adata_ids_sfaira.organism in adata.uns.keys():
                organism = adata.uns[self._adata_ids_sfaira.organism]
            else:
                # Declare as unknown organism and genome and make a group of its own:
                organism = adata.uns[self._adata_ids_sfaira.id]
            if isinstance(organism, list):
                if len(organism) == 1:
                    organism = organism[0]
                    assert isinstance(organism, str), organism
                else:
                    raise ValueError(f"tried to register mixed organism data set ({organism})")
            adata_id = adata.uns[self._adata_ids_sfaira.id]
            # Make up a new merged ID for data set indexing if there is a list of IDs in .uns.
            if isinstance(adata_id, list):
                adata_id = "_".join(adata_id)
            if organism not in adata_by_key.keys():
                adata_by_key[organism] = {}
                indices[organism] = {}
            try:
                adata_by_key[organism][adata_id] = adata
                indices[organism][adata_id] = np.arange(0, adata.n_obs)
            except TypeError as e:
                raise TypeError(f"{e} for {organism} or {adata.uns[self._adata_ids_sfaira.id]}")
        stores = dict([
            (k, StoreAnndata(adata_by_key=adata_by_key[k], indices=indices[k], in_memory=True))
            for k in adata_by_key.keys()
        ])
        super(StoresAnndata, self).__init__(stores=stores)


class StoresDao(StoreMultipleFeatureSpaceBase):

    _dataset_weights: Union[None, Dict[str, float]]

    def __init__(self,
                 cache_path: Union[str, os.PathLike, List[str], List[os.PathLike]],
                 columns: Union[None, List[str]] = None):
        """

        :param cache_path: Store directory.
        :param columns: Which columns to read into the obs copy in the output, see pandas.read_parquet().
        """
        # Collect all data loaders from files in directory:
        self._adata_ids_sfaira = AdataIdsSfaira()
        adata_by_key = {}
        x_by_key = {}
        indices = {}
        if not isinstance(cache_path, list) or isinstance(cache_path, tuple) or isinstance(cache_path, np.ndarray):
            cache_path = [cache_path]
        for cache_path_i in cache_path:
            for f in np.sort(os.listdir(cache_path_i)):
                adata = None
                x = None
                trial_path = os.path.join(cache_path_i, f)
                if os.path.isdir(trial_path):
                    # zarr-backed anndata are saved as directories with the elements of the array group as further sub
                    # directories, e.g. a directory called "X", and a file ".zgroup" which identifies the zarr group.
                    adata, x = read_dao(trial_path, use_dask=True, columns=columns, obs_separate=False, x_separate=True)
                if adata is not None:
                    organism = adata.uns[self._adata_ids_sfaira.organism]
                    if organism not in adata_by_key.keys():
                        adata_by_key[organism] = {}
                        x_by_key[organism] = {}
                        indices[organism] = {}
                    if adata.uns[self._adata_ids_sfaira.id] in adata_by_key[organism].keys():
                        print(f"WARNING: overwriting store entry in {adata.uns[self._adata_ids_sfaira.id]} in store "
                              f"{cache_path_i}.")
                    adata_by_key[organism][adata.uns[self._adata_ids_sfaira.id]] = adata
                    x_by_key[organism][adata.uns[self._adata_ids_sfaira.id]] = x
                    indices[organism][adata.uns[self._adata_ids_sfaira.id]] = np.arange(0, adata.n_obs)
        stores = dict([
            (k, StoreDao(adata_by_key=adata_by_key[k], x_by_key=x_by_key[k], indices=indices[k],
                         obs_by_key=None))
            for k in adata_by_key.keys()
        ])
        self._x_by_key = x_by_key
        super(StoresDao, self).__init__(stores=stores)


class StoresH5ad(StoreMultipleFeatureSpaceBase):

    def __init__(
            self,
            cache_path: Union[str, os.PathLike, List[str], List[os.PathLike]],
            in_memory: bool = False):
        # Collect all data loaders from files in directory:
        self._adata_ids_sfaira = AdataIdsSfaira()
        adata_by_key = {}
        indices = {}
        if not isinstance(cache_path, list) or isinstance(cache_path, tuple) or isinstance(cache_path, np.ndarray):
            cache_path = [cache_path]
        for cache_path_i in cache_path:
            for f in np.sort(os.listdir(cache_path_i)):
                adata = None
                trial_path = os.path.join(cache_path_i, f)
                if os.path.isfile(trial_path):
                    # Narrow down to supported file types:
                    if f.split(".")[-1] == "h5ad":
                        try:
                            adata = anndata.read_h5ad(
                                filename=trial_path,
                                backed="r" if in_memory else None,
                            )
                        except OSError as e:
                            adata = None
                            print(f"WARNING: for data set {f}: {e}")
                if adata is not None:
                    organism = adata.uns[self._adata_ids_sfaira.organism]
                    if organism not in adata_by_key.keys():
                        adata_by_key[organism] = {}
                        indices[organism] = {}
                    if adata.uns[self._adata_ids_sfaira.id] in adata_by_key[organism].keys():
                        print(f"WARNING: overwriting store entry in {adata.uns[self._adata_ids_sfaira.id]} in store "
                              f"{cache_path_i}.")
                    adata_by_key[organism][adata.uns[self._adata_ids_sfaira.id]] = adata
                    indices[organism][adata.uns[self._adata_ids_sfaira.id]] = np.arange(0, adata.n_obs)
        stores = dict([
            (k, StoreAnndata(adata_by_key=adata_by_key[k], indices=indices[k], in_memory=in_memory))
            for k in adata_by_key.keys()
        ])
        super(StoresH5ad, self).__init__(stores=stores)



import os
import pytest

from sfaira.consts import AdataIdsSfaira
from sfaira.data.store.io.io_dao import read_dao
from sfaira.unit_tests.data_for_tests.databases.utils import prepare_dsg_database
from sfaira.unit_tests.data_for_tests.databases.consts import CELLXGENE_DATASET_ID
from sfaira.unit_tests.data_for_tests.loaders import RELEASE_HUMAN, RELEASE_MOUSE
from sfaira.unit_tests.directories import DIR_DATABASE_STORE_DAO

MATCH_TO_RELEASE = {"Homo sapiens": RELEASE_HUMAN, "Mus musculus": RELEASE_MOUSE}


@pytest.mark.parametrize("database", [
    ("cellxgene", ["id", CELLXGENE_DATASET_ID]),
])
@pytest.mark.parametrize("subset_genes_to_type", [None, "protein_coding", ])
def test_streamline_features(database: str, subset_genes_to_type: str):
    database, subset_args = database
    dsg = prepare_dsg_database(database=database)
    dsg.subset(key=subset_args[0], values=subset_args[1])
    dsg.load()
    dsg.streamline_features(match_to_release=MATCH_TO_RELEASE, subset_genes_to_type=subset_genes_to_type)


@pytest.mark.parametrize("database", [
    ("cellxgene", ["id", CELLXGENE_DATASET_ID]),
])
@pytest.mark.parametrize("format", ["sfaira", ])
def test_streamline_metadata(database: str, format: str):
    database, subset_args = database
    dsg = prepare_dsg_database(database=database)
    dsg.subset(key=subset_args[0], values=subset_args[1])
    dsg.load()
    dsg.streamline_features(match_to_release=MATCH_TO_RELEASE, subset_genes_to_type="protein_coding")
    dsg.streamline_metadata(schema=format)
    adata = dsg.datasets[subset_args[1]].adata
    ids = AdataIdsSfaira()
    assert "CL:0000128" in adata.obs[ids.cell_type + ids.onto_id_suffix].values
    assert "oligodendrocyte" in adata.obs[ids.cell_type].values
    assert "MmusDv:0000061" in adata.obs[ids.development_stage + ids.onto_id_suffix].values
    assert "early adult stage" in adata.obs[ids.development_stage].values
    assert "UBERON:0002436" in adata.obs[ids.organ + ids.onto_id_suffix].values
    assert "primary visual cortex" in adata.obs[ids.organ].values


@pytest.mark.parametrize("store", ["dao", ])
@pytest.mark.parametrize("database", [
    ("cellxgene", ["id", CELLXGENE_DATASET_ID]),
])
def test_output_to_store(store: str, database: str):
    database, subset_args = database
    dsg = prepare_dsg_database(database=database)
    dsg.subset(key=subset_args[0], values=subset_args[1])
    dsg.load()
    dsg.streamline_features(match_to_release=MATCH_TO_RELEASE, subset_genes_to_type="protein_coding")
    dsg.streamline_metadata(schema="sfaira", clean_obs=True, clean_uns=True, clean_var=True, clean_obs_names=True,
                            keep_id_obs=True, keep_orginal_obs=False, keep_symbol_obs=True)
    dsg.write_distributed_store(dir_cache=DIR_DATABASE_STORE_DAO, store_format=store, dense=True)
    fn_store = os.path.join(DIR_DATABASE_STORE_DAO, subset_args[1])
    adata = read_dao(store=fn_store)
    ids = AdataIdsSfaira()
    assert "CL:0000128" in adata.obs[ids.cell_type + ids.onto_id_suffix].values
    assert "oligodendrocyte" in adata.obs[ids.cell_type].values
    assert "MmusDv:0000061" in adata.obs[ids.development_stage + ids.onto_id_suffix].values
    assert "early adult stage" in adata.obs[ids.development_stage].values



import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):
    """
    TODO: move state exact to diesase
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://www.ebi.ac.uk/arrayexpress/files/E-MTAB-5061/E-MTAB-5061.processed.1.zip"
        self.download_url_meta = "https://www.ebi.ac.uk/arrayexpress/files/E-MTAB-5061/E-MTAB-5061.sdrf.txt"

        self.assay_sc = "Smart-seq2"
        self.author = "Segerstolpe"
        self.doi_journal = "10.1016/j.cmet.2016.08.020"
        self.layer_counts = "X"
        self.organ = "pancreas"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2016

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.cell_type_obs_key = "Characteristics[cell type]"
        self.state_exact_obs_key = "Characteristics[disease]"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "E-MTAB-5061.processed.1.zip"),
        os.path.join(data_dir, "E-MTAB-5061.sdrf.txt")
    ]
    df = pd.read_csv(fn[0], sep="\t")
    df.index = df.index.get_level_values(0)
    df = df.drop("#samples", axis=1)
    df = df.T.iloc[:, :26178]
    adata = anndata.AnnData(df)
    adata.obs = pd.read_csv(fn[1], sep="\t").set_index("Source Name").loc[adata.obs.index]
    # filter observations which are not cells (empty wells, low quality cells etc.)
    adata = adata[adata.obs["Characteristics[cell type]"] != "not applicable"].copy()

    return adata



import numpy as np
import torch


class LossLoglikelihoodNb(torch.nn.Module):

    def __init__(self, average=True):
        super(LossLoglikelihoodNb, self).__init__()
        self.average = average

    def forward(self, preds, target):
        """Implements the negative log likelihood loss as VAE reconstruction loss"""
        x = target
        loc, scale = torch.chunk(preds, chunks=2, dim=1)

        eta_loc = torch.log(loc)
        eta_scale = torch.log(scale)

        log_r_plus_mu = torch.log(scale + loc)

        ll = torch.lgamma(scale + x)
        ll = ll - torch.lgamma(x + torch.ones_like(x))
        ll = ll - torch.lgamma(scale)
        ll = ll + torch.multiply(x, eta_loc - log_r_plus_mu) + torch.multiply(scale, eta_scale - log_r_plus_mu)

        ll = torch.clamp(ll, min=-300, max=300)
        neg_ll = -ll
        if self.average:
            neg_ll = torch.mean(neg_ll)
        else:
            # sum over features, average over batch
            neg_ll = neg_ll.sum(dim=1).sum(dim=1)
        return neg_ll


class LossLoglikelihoodGaussian(torch.nn.Module):

    def __init__(self, average=True):
        super(LossLoglikelihoodGaussian, self).__init__()
        self.average = average

    def forward(self, preds, target):
        """Implements the gaussian log likelihood loss as VAE reconstruction loss"""
        loc, scale = torch.chunk(preds, chunks=2, dim=1)

        ll = -torch.log(scale * torch.sqrt(2. * np.pi)) - 0.5 * torch.square((target - loc) / scale)
        ll = torch.clamp(ll, min=-300, max=300)
        neg_ll = -ll
        if self.average:
            neg_ll = torch.mean(neg_ll)
        else:
            # sum over features, average over batch
            neg_ll = torch.mean(torch.sum(neg_ll, dim=1), dim=0)
        return neg_ll


class LossCrossentropyAgg(torch.nn.Module):

    def __init__(self):
        super(LossCrossentropyAgg, self).__init__()

    def forward(self, preds, target):
        """ Modified crossentropy that aggregates allowed output classes into single class. """
        preds = torch.clamp(preds, min=1e-10, max=1.)
        ll_cce_agg = -torch.log(torch.mean(target * preds, dim=1, keepdim=False))
        return ll_cce_agg


class KLLoss(torch.nn.Module):

    def __init__(self):
        super(KLLoss, self).__init__()
        self.beta = self.register_buffer('beta', torch.Tensor(1.))

    def forward(self, preds, target):
        expected_logqz_x, expected_logpz = torch.chunk(preds, chunks=2, dim=1)

        kl_loss = torch.mean(expected_logqz_x - expected_logpz, dim=0)
        return self.beta * kl_loss



import abc
try:
    import tensorflow as tf
except ImportError:
    tf = None
from sfaira.models.base import BasicModelKeras


class BasicModelEmbedding:

    @abc.abstractmethod
    def predict(self, x, **kwargs):
        pass


class BasicModelKerasCelltype(BasicModelKeras):
    """
    This base class defines model attributes shared across all tf.keras cell type models.
    """

    def predict(self, x, **kwarg):
        return self.training_model.predict(x)



import logging

import os
from rich import print
from typing import Dict

from sfaira.commands.utils import get_ds
from sfaira.consts.utils import clean_doi
from sfaira.data import DatasetBase

log = logging.getLogger(__name__)


class H5adExport:

    datasets: Dict[str, DatasetBase]
    doi: str
    doi_sfaira_repr: str
    path_cache: str
    path_data: str
    path_loader: str
    path_out: str
    schema: str

    def __init__(self, doi, path_cache, path_data, path_loader, path_out, schema):
        self.doi = doi
        self.doi_sfaira_repr = clean_doi(self.doi)
        self.path_cache = path_cache
        self.path_data = path_data
        self.path_loader = path_loader
        self.path_out = path_out
        if schema not in ["cellxgene"]:
            raise ValueError(f"Did not recognize schema {schema}")
        self.schema = schema

    def write(self):
        self._load_objects()
        self._write_h5ads()

    def _load_objects(self):
        dsg, _ = get_ds(doi_sfaira_repr=self.doi_sfaira_repr, path_cache=self.path_cache, path_data=self.path_data,
                        path_loader=self.path_loader)
        dsg.load(load_raw=False, allow_caching=True)
        if self.schema == "cellxgene":
            dsg.streamline_features(schema="cellxgene:" + "2.0.0")
            dsg.streamline_metadata(
                schema=self.schema.lower(),
                clean_obs=False,
                clean_var=False,
                clean_uns=True,
                clean_obs_names=False,
                keep_orginal_obs=False,
                keep_symbol_obs=True,
                keep_id_obs=True,
            )
            dsg.collapse_counts()
        self.datasets = dsg.datasets

    def _write_h5ads(self):
        counter = 0
        for k, v in self.datasets.items():
            fn = v.doi_cleaned_id + ".h5ad"
            dir_name = v.directory_formatted_doi
            if not os.path.exists(os.path.join(self.path_out, dir_name)):
                os.makedirs(os.path.join(self.path_out, dir_name))
            fn_out = os.path.join(self.path_out, dir_name, fn)
            print(f'[bold orange]Sfaira butler: "Preparing {fn_out} for you."')
            v.adata.write_h5ad(fn_out)
            counter += 1
        print(f'[bold orange]Sfaira butler: "I wrote a total of {counter} .h5ad files."')



import os
import re
import pydoc
import shutil
import sys

from sfaira.commands.consts import PACKAGE_LOADER_PATH, PACKAGE_SFAIRAE_LOADER_PATH
from sfaira.data import DatasetGroupDirectoryOriented

try:
    import sfaira_extension as sfairae
except ImportError:
    sfairae = None


def doi_lint(x: str) -> bool:
    """
    Check if string matches DOI format.

    Allowed formats are:
        - an actual DOI startin with "10*"
        - a string starting with "no_doi*" indicating that a dataset without corresponding DOI is used
        - a sfaira formated DOI starting with "d10*" or "dno_doi*"
    """
    return re.match(r'\b10\.\d+/[\w.]+\b', x) or x.startswith("no_doi") or \
        x.startswith("d10_") or x.startswith("dno_doi")


def get_pydoc(path_loader, doi_sfaira_repr) -> [str, str]:
    pydoc_handle_sfaira = "sfaira.data.dataloaders.loaders"
    file_path_sfaira = os.path.dirname(str(pydoc.locate(pydoc_handle_sfaira + ".FILE_PATH")))
    pydoc_handle_sfairae = "sfaira_extension.data.dataloaders.loaders" if sfairae else None
    file_path_sfairae = os.path.dirname(str(pydoc.locate(pydoc_handle_sfairae + ".FILE_PATH"))) if sfairae else None

    # Check if loader name is a directory either in external, sfaira or sfaira_extension loader collections:
    if path_loader != PACKAGE_LOADER_PATH and path_loader != PACKAGE_SFAIRAE_LOADER_PATH:  # external
        if doi_sfaira_repr in os.listdir(path_loader):
            if path_loader not in sys.path:
                sys.path.append(path_loader)
            pydoc_handle = doi_sfaira_repr
            # Imitate result from pydoc locate:
            file_path = os.path.join(path_loader, doi_sfaira_repr, "__init__.py")
        else:
            raise ValueError(f"did not find loader {doi_sfaira_repr} in custom path {os.listdir(path_loader)}")
    elif doi_sfaira_repr in os.listdir(file_path_sfaira):
        pydoc_handle = pydoc_handle_sfaira + "." + doi_sfaira_repr
        file_path = os.path.join(file_path_sfaira, doi_sfaira_repr, "__init__.py")
    elif file_path_sfairae and doi_sfaira_repr in os.listdir(file_path_sfairae):
        pydoc_handle = pydoc_handle_sfairae + "." + doi_sfaira_repr
        file_path = os.path.join(file_path_sfairae, doi_sfaira_repr, "__init__.py")
    else:
        raise ValueError("data loader not found in sfaira and also not in sfaira_extension")
    return file_path, pydoc_handle


def get_ds(doi_sfaira_repr, path_data, path_loader, path_cache=None, clear_cache=True):
    file_path, _ = get_pydoc(path_loader=path_loader, doi_sfaira_repr=doi_sfaira_repr)
    if path_cache is not None and clear_cache:
        shutil.rmtree(path_cache, ignore_errors=True)

    ds = DatasetGroupDirectoryOriented(
        file_base=file_path,
        data_path=path_data,
        meta_path=None,
        cache_path=path_cache,
    )

    return ds, path_cache
